from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker
from pylint.lint import PyLinter


class NoDbQueryStringFormatChecker(BaseChecker):
    __implements__ = IAstroidChecker

    name = "no-database-query-string-format"
    priority = -1
    msgs = {
        "W9000": (
            "Python string formatting found in database query. Database queries should use native parameter substitution.",
            "database-query-string-format",
            "This allows CQL/SQL injection.",
        )
    }

    def __init__(self, linter: PyLinter = None):
        super().__init__(linter)
        self.string_sub_queries: set = set()
        self.query_verbs = {"select", "update", "delete", "insertinto", "truncate"}

    def check_string_is_query(self, string: str) -> bool:
        query = string.split(" ")
        if query[0].lower() in self.query_verbs and ("{}" in query or "%s" in query):
            return True
        return False

    def visit_assign(self, node: nodes) -> None:
        """Check variables with queries using string formatting."""
        variable_is_query = False
        if (
            isinstance(node.value, nodes.BinOp)
            and node.value.op == "%"
            and isinstance(node.value.left, nodes.Const)
            and self.check_string_is_query(node.value.left.value)
        ):
            variable_is_query = True
        elif (
            isinstance(node.value, nodes.Call)
            and isinstance(node.targets[0], nodes.Name)
            and isinstance(node.value.func, nodes.Attribute)
            and node.value.func.attrname == "format"
            and self.check_string_is_query(node.value.func.expr.value)
        ):
            variable_is_query = True

        if variable_is_query and node.targets[0].name not in self.string_sub_queries:
            self.string_sub_queries.add(node.targets[0].name)
        elif not variable_is_query and node.targets[0].name in self.string_sub_queries:
            self.string_sub_queries.remove(node.targets[0].name)

    def visit_call(self, node: nodes) -> None:
        """Check whether execute calls have queries using string formatting."""
        if isinstance(node.func, nodes.Attribute) and node.func.attrname == "execute" and node.args:
            if (
                isinstance(node.args[0], nodes.Name)
                and node.args[0].name in self.string_sub_queries
            ):
                self.add_message("database-query-string-format", node=node)
            elif (
                isinstance(node.args[0], nodes.BinOp)
                and node.args[0].op == "%"
                and isinstance(node.args[0].left, nodes.Const)
                and self.check_string_is_query(node.args[0].left.value)
            ):
                self.add_message("database-query-string-format", node=node)
            elif (
                isinstance(node.args[0], nodes.Call)
                and isinstance(node.args[0].func, nodes.Attribute)
                and node.args[0].func.attrname == "format"
                and self.check_string_is_query(node.args[0].func.expr.value)
            ):
                self.add_message("database-query-string-format", node=node)

    def leave_module(self, node: nodes) -> nodes:
        self.string_sub_queries = set()
        return node

    def leave_classdef(self, node: nodes) -> nodes:
        self.string_sub_queries = set()
        return node

    def leave_functiondef(self, node: nodes) -> nodes:
        self.string_sub_queries = set()
        return node



import unittest

from unittest import mock

from baseplate import Baseplate
from baseplate import BaseplateObserver
from baseplate import LocalSpan
from baseplate import ParentSpanAlreadyFinishedError
from baseplate import RequestContext
from baseplate import ReusedContextObjectError
from baseplate import ServerSpan
from baseplate import ServerSpanObserver
from baseplate import Span
from baseplate import SpanObserver
from baseplate import TraceInfo
from baseplate.clients import ContextFactory
from baseplate.lib import config


def make_test_server_span(context=None):
    if not context:
        context = mock.Mock()
    return ServerSpan(1, 2, 3, None, 0, "name", context)


def make_test_span(context=None, local=False):
    if not context:
        context = mock.Mock()
    span = Span(1, 2, 3, None, 0, "name", context)
    if local:
        span = LocalSpan(1, 2, 3, None, 0, "name", context)
    return span


class BaseplateTests(unittest.TestCase):
    def test_server_observer_made(self):
        baseplate = Baseplate()
        mock_context = baseplate.make_context_object()
        mock_observer = mock.Mock(spec=BaseplateObserver)
        baseplate.register(mock_observer)
        server_span = baseplate.make_server_span(mock_context, "name", TraceInfo(1, 2, 3, None, 0))

        self.assertEqual(baseplate.observers, [mock_observer])
        self.assertEqual(mock_observer.on_server_span_created.call_count, 1)
        self.assertEqual(
            mock_observer.on_server_span_created.call_args, mock.call(mock_context, server_span)
        )

    def test_null_server_observer(self):
        baseplate = Baseplate()
        mock_context = baseplate.make_context_object()
        mock_observer = mock.Mock(spec=BaseplateObserver)
        mock_observer.on_server_span_created.return_value = None
        baseplate.register(mock_observer)
        server_span = baseplate.make_server_span(mock_context, "name", TraceInfo(1, 2, 3, None, 0))

        self.assertEqual(server_span.observers, [])

    def test_configure_context_supports_complex_specs(self):
        from baseplate.clients.thrift import ThriftClient
        from baseplate.thrift import BaseplateServiceV2

        app_config = {
            "enable_some_fancy_feature": "true",
            "thrift.foo.endpoint": "localhost:9090",
            "thrift.bar.endpoint": "localhost:9091",
        }

        baseplate = Baseplate(app_config)
        baseplate.configure_context(
            {
                "enable_some_fancy_feature": config.Boolean,
                "thrift": {
                    "foo": ThriftClient(BaseplateServiceV2.Client),
                    "bar": ThriftClient(BaseplateServiceV2.Client),
                },
            },
        )

        context = baseplate.make_context_object()
        with baseplate.make_server_span(context, "test"):
            self.assertTrue(context.enable_some_fancy_feature)
            self.assertIsNotNone(context.thrift.foo)
            self.assertIsNotNone(context.thrift.bar)

    def test_with_server_context(self):
        baseplate = Baseplate()
        observer = mock.Mock(spec=BaseplateObserver)
        baseplate.register(observer)

        observer.on_server_span_created.assert_not_called()
        with baseplate.server_context("example") as context:
            observer.on_server_span_created.assert_called_once()
            self.assertIsInstance(context, RequestContext)

    def test_add_to_context(self):
        baseplate = Baseplate()
        forty_two_factory = mock.Mock(spec=ContextFactory)
        forty_two_factory.make_object_for_context = mock.Mock(return_value=42)
        baseplate.add_to_context("forty_two", forty_two_factory)
        baseplate.add_to_context("true", True)

        context = baseplate.make_context_object()

        self.assertEqual(42, context.forty_two)
        self.assertTrue(context.true)

    def test_add_to_context_supports_complex_specs(self):
        baseplate = Baseplate()
        forty_two_factory = mock.Mock(spec=ContextFactory)
        forty_two_factory.make_object_for_context = mock.Mock(return_value=42)
        context_spec = {
            "forty_two": forty_two_factory,
            "true": True,
            "nested": {"foo": "bar"},
        }
        baseplate.add_to_context("complex", context_spec)

        context = baseplate.make_context_object()

        self.assertEqual(42, context.complex.forty_two)
        self.assertTrue(context.complex.true)
        self.assertEqual("bar", context.complex.nested.foo)


class SpanTests(unittest.TestCase):
    def test_events(self):
        mock_observer = mock.Mock(spec=SpanObserver)

        span = make_test_span()
        span.register(mock_observer)

        span.start()
        self.assertEqual(mock_observer.on_start.call_count, 1)

        span.set_tag("key", "value")
        mock_observer.on_set_tag("key", "value")

        span.log("name", "payload")
        mock_observer.on_log("name", "payload")

        span.finish()
        mock_observer.on_finish(exc_info=None)

    def test_context(self):
        mock_observer = mock.Mock(spec=SpanObserver)

        span = make_test_span()
        span.register(mock_observer)

        with span:
            self.assertEqual(mock_observer.on_start.call_count, 1)
        self.assertEqual(mock_observer.on_finish.call_count, 1)

    def test_context_with_tags(self):
        mock_observer = mock.Mock(spec=SpanObserver)

        span = make_test_span()
        span.register(mock_observer)

        tags = {
            "k1": "v1",
            "k2": "v2",
        }
        with span.with_tags(tags) as span:
            self.assertEqual(mock_observer.on_start.call_count, 1)
            mock_observer.on_set_tag.assert_has_calls(
                [
                    mock.call("k1", "v1"),
                    mock.call("k2", "v2"),
                ]
            )
        self.assertEqual(mock_observer.on_finish.call_count, 1)

    def test_context_with_exception(self):
        mock_observer = mock.Mock(spec=SpanObserver)

        # span = Span(1, 2, 3, None, 0, "name")
        span = make_test_span()
        span.register(mock_observer)

        class TestException(Exception):
            pass

        exc = TestException()
        with self.assertRaises(TestException):
            with span:
                raise exc
        self.assertEqual(mock_observer.on_finish.call_count, 1)
        _, captured_exc, _ = mock_observer.on_finish.call_args[0][0]
        self.assertEqual(captured_exc, exc)


class ServerSpanTests(unittest.TestCase):
    @mock.patch("random.getrandbits", autospec=True)
    def test_make_child(self, mock_getrandbits):
        mock_getrandbits.return_value = 0xCAFE

        mock_observer = mock.Mock(spec=ServerSpanObserver)
        mock_context = mock.Mock()

        server_span = ServerSpan("trace", "parent", "id", None, 0, "name", mock_context)
        server_span.register(mock_observer)
        child_span = server_span.make_child("child_name")

        self.assertEqual(child_span.name, "child_name")
        self.assertEqual(child_span.id, "51966")
        self.assertEqual(child_span.trace_id, "trace")
        self.assertEqual(child_span.parent_id, "id")
        self.assertEqual(mock_observer.on_child_span_created.call_count, 1)
        self.assertEqual(mock_observer.on_child_span_created.call_args, mock.call(child_span))

    def test_null_child(self):
        mock_observer = mock.Mock(spec=ServerSpanObserver)
        mock_observer.on_child_span_created.return_value = None

        server_span = make_test_server_span()
        # server_span = ServerSpan("trace", "parent", "id", None, 0, "name")
        server_span.register(mock_observer)
        child_span = server_span.make_child("child_name")

        self.assertEqual(child_span.observers, [])

    @mock.patch("random.getrandbits", autospec=True)
    def test_make_local_span(self, mock_getrandbits):
        mock_getrandbits.return_value = 0xCAFE
        mock_observer = mock.Mock(spec=ServerSpanObserver)
        mock_context = mock.Mock()
        mock_cloned_context = mock.Mock()
        mock_context.clone.return_value = mock_cloned_context

        server_span = ServerSpan("trace", "parent", "id", None, 0, "name", mock_context)
        server_span.register(mock_observer)
        local_span = server_span.make_child("test_op", local=True, component_name="test_component")

        self.assertEqual(local_span.name, "test_op")
        self.assertEqual(local_span.id, "51966")
        self.assertEqual(local_span.trace_id, "trace")
        self.assertEqual(local_span.parent_id, "id")
        self.assertEqual(mock_observer.on_child_span_created.call_count, 1)
        self.assertEqual(mock_observer.on_child_span_created.call_args, mock.call(local_span))

    @mock.patch("random.getrandbits", autospec=True)
    def test_make_local_span_copies_context(self, mock_getrandbits):
        mock_getrandbits.return_value = 0xCAFE
        mock_observer = mock.Mock(spec=ServerSpanObserver)
        mock_context = mock.Mock()
        mock_cloned_context = mock.Mock()
        mock_context.clone.return_value = mock_cloned_context

        server_span = ServerSpan("trace", "parent", "id", None, 0, "name", mock_context)
        server_span.register(mock_observer)
        local_span = server_span.make_child("test_op", local=True, component_name="test_component")
        self.assertNotEqual(local_span.context, mock_context)


class LocalSpanTests(unittest.TestCase):
    def test_events(self):
        mock_observer = mock.Mock(spec=SpanObserver)

        span = make_test_span(local=True)
        span.register(mock_observer)

        span.start()
        self.assertEqual(mock_observer.on_start.call_count, 1)

        span.set_tag("key", "value")
        mock_observer.on_set_tag("key", "value")

        span.log("name", "payload")
        mock_observer.on_log("name", "payload")

        span.make_child("local_child")
        self.assertEqual(mock_observer.on_child_span_created.call_count, 1)
        span.finish()
        mock_observer.on_finish(exc_info=None)

    @mock.patch("random.getrandbits", autospec=True)
    def test_make_child(self, mock_getrandbits):
        mock_getrandbits.return_value = 0xCAFE

        mock_observer = mock.Mock(spec=SpanObserver)
        mock_context = mock.Mock()

        local_span = LocalSpan("trace", "parent", "id", None, 0, "name", mock_context)
        local_span.register(mock_observer)
        child_span = local_span.make_child("child_name")

        self.assertEqual(child_span.name, "child_name")
        self.assertEqual(child_span.id, "51966")
        self.assertEqual(child_span.trace_id, "trace")
        self.assertEqual(child_span.parent_id, "id")
        self.assertEqual(mock_observer.on_child_span_created.call_count, 1)
        self.assertEqual(mock_observer.on_child_span_created.call_args, mock.call(child_span))

    def test_context_object_reused(self):
        baseplate = Baseplate()
        context = baseplate.make_context_object()

        with baseplate.make_server_span(context, "foo"):
            pass

        with self.assertRaises(ReusedContextObjectError):
            with baseplate.make_server_span(context, "bar"):
                pass

    def test_parent_already_finished(self):
        mock_context = mock.Mock()
        local_span = LocalSpan("trace", "parent", "id", None, 0, "name", mock_context)
        local_span.start()
        local_span.finish()

        with self.assertRaises(ParentSpanAlreadyFinishedError):
            local_span.make_child("foo")


class TraceInfoTests(unittest.TestCase):
    def test_new_does_not_have_parent_id(self):
        new_trace_info = TraceInfo.new()
        self.assertIsNone(new_trace_info.parent_id)

    def test_new_does_not_set_flags(self):
        new_trace_info = TraceInfo.new()
        self.assertIsNone(new_trace_info.flags)

    def test_new_does_not_set_sampled(self):
        new_trace_info = TraceInfo.new()
        self.assertIsNone(new_trace_info.sampled)

    def test_from_upstream_fails_on_invalid_sampled(self):
        with self.assertRaises(ValueError) as e:
            TraceInfo.from_upstream(1, 2, 3, "True", None)
        self.assertEqual(str(e.exception), "invalid sampled value")

    def test_from_upstream_fails_on_invalid_flags(self):
        with self.assertRaises(ValueError) as e:
            TraceInfo.from_upstream(1, 2, 3, True, -1)
        self.assertEqual(str(e.exception), "invalid flags value")

    def test_from_upstream_handles_no_sampled_or_flags(self):
        span = TraceInfo.from_upstream(1, 2, 3, None, None)
        self.assertIsNone(span.sampled)
        self.assertIsNone(span.flags)



import gevent.monkey
import pytest

from baseplate import Baseplate
from baseplate.observers.timeout import ServerTimeout
from baseplate.observers.timeout import TimeoutBaseplateObserver


def _create_baseplate_object(timeout: str):
    baseplate = Baseplate()

    observer = TimeoutBaseplateObserver.from_config({"server_timeout.default": timeout})
    baseplate.register(observer)
    return baseplate


def test_default_timeout():
    baseplate = _create_baseplate_object("50 milliseconds")
    context = baseplate.make_context_object()
    with baseplate.make_server_span(context, "test"):
        with pytest.raises(ServerTimeout):
            gevent.sleep(1)

    context = baseplate.make_context_object()
    with baseplate.make_server_span(context, "test"):
        gevent.sleep(0)  # shouldn't time out since it's so fast!


def test_route_specific_timeout():
    baseplate = _create_baseplate_object("5 milliseconds")
    context = baseplate.make_context_object()
    with baseplate.make_server_span(context, "test"):
        with pytest.raises(ServerTimeout):
            gevent.sleep(1)

    context = baseplate.make_context_object()
    with baseplate.make_server_span(context, "test"):
        gevent.sleep(0)


def test_timeout_from_context():
    baseplate = _create_baseplate_object("1 hour")
    baseplate.add_to_context("deadline_budget", 0.01)

    # tests short deadline_budget causes timeout
    context = baseplate.make_context_object()
    with baseplate.make_server_span(context, "test"):
        with pytest.raises(ServerTimeout):
            gevent.sleep(1)

    # tests long deadline_budget and server timeout doesn't timeout
    context = baseplate.make_context_object()
    baseplate.add_to_context("deadline_budget", 1000)
    with baseplate.make_server_span(context, "test"):
        gevent.sleep(0.01)  # shouldn't time out since timeouts are so long


def test_pool_timeout_with_long_deadline_budget():
    baseplate = _create_baseplate_object("5 milliseconds")
    baseplate.add_to_context("deadline_budget", 1000)

    context = baseplate.make_context_object()
    with baseplate.make_server_span(context, "test"):
        with pytest.raises(ServerTimeout):
            gevent.sleep(0.01)



import os
import unittest

from unittest import mock

import pytest

from prometheus_client import REGISTRY

try:
    import redis
except ImportError:
    raise unittest.SkipTest("redis-py is not installed")
else:
    del redis
from redis.exceptions import ConnectionError

from baseplate.lib.config import ConfigurationError
from baseplate.clients.redis import pool_from_config
from baseplate.clients.redis import ACTIVE_REQUESTS
from baseplate.clients.redis import REQUESTS_TOTAL
from baseplate.clients.redis import LATENCY_SECONDS
from baseplate.clients.redis import MonitoredRedisConnection


class DummyConnection:
    description_format = "DummyConnection<>"

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.pid = os.getpid()

    def connect(self):
        pass

    def can_read(self):
        return False

    def send_command(self, command):
        pass

    def read_response(self):
        # Must return same number as test_pipeline_instrumentation calls
        return ["OK", "OK"]

    def pack_commands(self, *args):
        pass

    def send_packed_command(self, *args):
        pass

    def disconnect(self):
        pass


class TestMonitoredRedisConnection:
    def setup(self):
        ACTIVE_REQUESTS.clear()
        REQUESTS_TOTAL.clear()
        LATENCY_SECONDS.clear()

    @pytest.fixture
    def app_config(self):
        yield {
            "redis.url": "redis://localhost:1234/0",
        }

    @pytest.fixture
    def expected_labels(self):
        yield {
            "redis_client_name": "test_client",
            "redis_type": "standalone",
            "redis_command": "some_command",
            "redis_database": "0",
        }

    @pytest.fixture
    def connection(self):
        yield {"connection_class": DummyConnection}

    @pytest.fixture
    def connection_pool(self, app_config, connection):
        yield pool_from_config(app_config=app_config, prefix="redis.", **connection)

    @pytest.fixture
    def context(self):
        yield mock.MagicMock()

    @pytest.fixture
    def span(self):
        yield mock.MagicMock()

    @pytest.fixture
    def monitored_redis_connection(self, span, connection_pool):
        yield MonitoredRedisConnection(
            "redis_context_name", span, connection_pool, redis_client_name="test_client"
        )

    def test_execute_command_exc_redis_err(
        self, monitored_redis_connection, expected_labels, app_config
    ):
        monitored_redis_connection.connection_pool = pool_from_config(
            app_config=app_config, client_name="test_client"
        )
        with pytest.raises(ConnectionError):  # ConnectionError inherits from RedisError
            monitored_redis_connection.execute_command("some_command")
        assert REGISTRY.get_sample_value(f"{ACTIVE_REQUESTS._name}", expected_labels) == 0
        expected_labels["redis_success"] = "false"
        assert (
            REGISTRY.get_sample_value(
                f"{LATENCY_SECONDS._name}_bucket", {**expected_labels, **{"le": "+Inf"}}
            )
            == 1
        )
        assert REGISTRY.get_sample_value(f"{REQUESTS_TOTAL._name}_total", expected_labels) == 1

    def test_execute_command(self, monitored_redis_connection, expected_labels):
        monitored_redis_connection.execute_command("some_command")
        # assert [i for i in REGISTRY.collect()] == ""
        assert REGISTRY.get_sample_value(f"{ACTIVE_REQUESTS._name}", expected_labels) == 0
        expected_labels["redis_success"] = "true"
        assert (
            REGISTRY.get_sample_value(
                f"{LATENCY_SECONDS._name}_bucket", {**expected_labels, **{"le": "+Inf"}}
            )
            == 1
        )
        assert REGISTRY.get_sample_value(f"{REQUESTS_TOTAL._name}_total", expected_labels) == 1

    def test_pipeline_instrumentation(self, monitored_redis_connection, expected_labels):
        active_labels = {**expected_labels, "redis_command": "pipeline"}
        mock_manager = mock.Mock()
        with mock.patch.object(
            ACTIVE_REQUESTS.labels(**active_labels),
            "inc",
            wraps=ACTIVE_REQUESTS.labels(**active_labels).inc,
        ) as active_inc_spy_method:
            mock_manager.attach_mock(active_inc_spy_method, "inc")
            with mock.patch.object(
                ACTIVE_REQUESTS.labels(**active_labels),
                "dec",
                wraps=ACTIVE_REQUESTS.labels(**active_labels).dec,
            ) as active_dec_spy_method:
                mock_manager.attach_mock(active_dec_spy_method, "dec")

                monitored_redis_connection.pipeline("test").set("hello", 42).set(
                    "goodbye", 23
                ).execute()
                labels = {**active_labels, "redis_success": "true"}
                assert (
                    REGISTRY.get_sample_value(f"{REQUESTS_TOTAL._name}_total", labels) == 1.0
                ), "Unexpected value for REQUESTS_TOTAL metric. Expected one 'pipeline' command"
                assert (
                    REGISTRY.get_sample_value(
                        f"{LATENCY_SECONDS._name}_bucket", {**labels, "le": "+Inf"}
                    )
                    == 1.0
                ), "Expected one 'pipeline' latency request"
                assert mock_manager.mock_calls == [
                    mock.call.inc(),
                    mock.call.dec(),
                ], "Instrumentation should increment and then decrement active requests exactly once"
                assert (
                    REGISTRY.get_sample_value(ACTIVE_REQUESTS._name, active_labels) == 0.0
                ), "Should have 0 (and not None) active requests"

    def test_pipeline_instrumentation_failing(
        self,
        monitored_redis_connection,
        expected_labels,
        app_config,
        redis_client_name="test_client",
    ):
        active_labels = {**expected_labels, "redis_command": "pipeline"}
        mock_manager = mock.Mock()
        with mock.patch.object(
            ACTIVE_REQUESTS.labels(**active_labels),
            "inc",
            wraps=ACTIVE_REQUESTS.labels(**active_labels).inc,
        ) as active_inc_spy_method:
            mock_manager.attach_mock(active_inc_spy_method, "inc")
            with mock.patch.object(
                ACTIVE_REQUESTS.labels(**active_labels),
                "dec",
                wraps=ACTIVE_REQUESTS.labels(**active_labels).dec,
            ) as active_dec_spy_method:
                mock_manager.attach_mock(active_dec_spy_method, "dec")

                monitored_redis_connection.connection_pool = pool_from_config(
                    app_config=app_config, client_name="test_client"
                )
                with pytest.raises(ConnectionError):
                    monitored_redis_connection.pipeline("test").set("hello", 42).set(
                        "goodbye", 23
                    ).execute()
                labels = {**active_labels, "redis_success": "false"}
                assert (
                    REGISTRY.get_sample_value(f"{REQUESTS_TOTAL._name}_total", labels) == 1.0
                ), "Unexpected value for REQUESTS_TOTAL metric. Expected one 'pipeline' command"
                assert (
                    REGISTRY.get_sample_value(
                        f"{LATENCY_SECONDS._name}_bucket", {**labels, "le": "+Inf"}
                    )
                    == 1.0
                ), "Expected one 'pipeline' latency request"
                assert mock_manager.mock_calls == [
                    mock.call.inc(),
                    mock.call.dec(),
                ], "Instrumentation should increment and then decrement active requests exactly once"
                assert (
                    REGISTRY.get_sample_value(ACTIVE_REQUESTS._name, active_labels) == 0.0
                ), "Should have 0 (and not None) active requests"


class TestPoolFromConfig:
    def test_empty_config(self):
        with pytest.raises(ConfigurationError):
            pool_from_config({})

    def test_basic_url(self):
        pool = pool_from_config({"redis.url": "redis://localhost:1234/0"})

        assert pool.connection_kwargs["host"] == "localhost"
        assert pool.connection_kwargs["port"] == 1234
        assert pool.connection_kwargs["db"] == 0

    def test_timeouts(self):
        pool = pool_from_config(
            {
                "redis.url": "redis://localhost:1234/0",
                "redis.socket_timeout": "30 seconds",
                "redis.socket_connect_timeout": "300 milliseconds",
            }
        )

        assert pool.connection_kwargs["socket_timeout"] == 30
        assert pool.connection_kwargs["socket_connect_timeout"] == 0.3

    def test_kwargs_passthrough(self):
        pool = pool_from_config({"redis.url": "redis://localhost:1234/0"}, example="present")

        assert pool.connection_kwargs["example"] == "present"

    def test_alternate_prefix(self):
        pool_from_config({"noodle.url": "redis://localhost:1234/0"}, prefix="noodle.")



import typing

from typing import Tuple
from typing import Type
from typing import Union

from baseplate.lib.file_watcher import _NOT_LOADED
from baseplate.lib.file_watcher import FileWatcher
from baseplate.lib.file_watcher import T
from baseplate.lib.file_watcher import WatchedFileNotAvailableError


class FakeFileWatcher(FileWatcher):
    """Fake file watcher for testing purposes.

    Use this in place of a :py:class:`~baseplate.lib.file_watcher.FileWatcher`
    in tests to avoid having to load an actual file:

    .. testsetup::

        from baseplate.testing.lib.file_watcher import FakeFileWatcher

    .. doctest::

        >>> file_watcher = FakeFileWatcher()
        >>> file_watcher.get_data()
        Traceback (most recent call last):
        baseplate.lib.file_watcher.WatchedFileNotAvailableError: /fake-file-watcher: no value set
        >>> file_watcher.data = "test"
        >>> file_watcher.get_data()
        'test'

    .. versionadded:: 1.5

    """

    # pylint: disable=super-init-not-called
    def __init__(self, data: Union[T, Type[_NOT_LOADED]] = _NOT_LOADED, mtime: float = 1234):
        self.data = data
        self.mtime = mtime

    def get_data_and_mtime(self) -> Tuple[T, float]:
        if self.data is _NOT_LOADED:
            raise WatchedFileNotAvailableError("/fake-file-watcher", Exception("no value set"))
        return typing.cast(T, self.data), self.mtime



from random import random
from typing import Any
from typing import Dict
from typing import Optional
from typing import Set

from baseplate import _ExcInfo
from baseplate import BaseplateObserver
from baseplate import LocalSpan
from baseplate import RequestContext
from baseplate import Span
from baseplate import SpanObserver
from baseplate.lib import config
from baseplate.lib import metrics


class TaggedMetricsBaseplateObserver(BaseplateObserver):
    """Metrics collecting observer.

    This observer reports metrics to statsd in the Influx StatsD format. It does three important things:

    * it tracks the time taken in serving each request.
    * it batches all metrics generated during a request into as few packets
      as possible.
    * it adds tags to the metric if they are in the config tag allowlist

    The batch is accessible to your application during requests as the
    ``metrics`` attribute on the :py:class:`~baseplate.RequestContext`.

    :param client: The client where metrics will be sent.
    :param cfg: the parsed application config with the tag allowlist

    """

    def __init__(self, client: metrics.Client, allowlist: Set[str], sample_rate: float = 1.0):
        self.client = client
        self.allowlist = allowlist
        self.sample_rate = sample_rate

    @classmethod
    def from_config_and_client(
        cls, raw_config: config.RawConfig, client: metrics.Client
    ) -> "TaggedMetricsBaseplateObserver":
        cfg = config.parse_config(
            raw_config,
            {
                "metrics": {
                    "allowlist": config.Optional(config.TupleOf(config.String), default=[]),
                },
                "metrics_observer": {"sample_rate": config.Optional(config.Percent, default=1.0)},
            },
        )
        return cls(
            client,
            allowlist=set(cfg.metrics.allowlist) | {"client", "endpoint"},
            sample_rate=cfg.metrics_observer.sample_rate,
        )

    def on_server_span_created(self, context: RequestContext, server_span: Span) -> None:
        batch = self.client.batch()
        context.metrics = batch
        if self.sample_rate == 1.0 or random() < self.sample_rate:
            observer: SpanObserver = TaggedMetricsServerSpanObserver(
                batch, server_span, self.allowlist, self.sample_rate
            )
        else:
            observer = TaggedMetricsServerSpanDummyObserver(batch)
        server_span.register(observer)


class TaggedMetricsServerSpanDummyObserver(SpanObserver):
    # for requests that aren't sampled
    def __init__(self, batch: metrics.Batch):
        self.batch = batch

    def on_start(self) -> None:
        pass

    def on_incr_tag(self, key: str, delta: float) -> None:
        pass

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        self.batch.flush()

    def on_child_span_created(self, span: Span) -> None:
        pass


class TaggedMetricsServerSpanObserver(SpanObserver):
    def __init__(
        self, batch: metrics.Batch, server_span: Span, allowlist: Set[str], sample_rate: float = 1.0
    ):
        self.batch = batch
        self.span = server_span
        self.base_name = "baseplate.server"
        self.allowlist = allowlist
        self.tags: Dict[str, Any] = {}
        self.timer = batch.timer(f"{self.base_name}.latency")
        self.counters: Dict[str, float] = {}
        self.sample_rate = sample_rate

    def on_start(self) -> None:
        self.tags["endpoint"] = self.span.name
        self.timer.start(self.sample_rate)

    def on_incr_tag(self, key: str, delta: float) -> None:
        self.counters[key] = delta

    def on_set_tag(self, key: str, value: Any) -> None:
        self.tags[key] = value

    def on_child_span_created(self, span: Span) -> None:
        observer: SpanObserver
        if isinstance(span, LocalSpan):
            observer = TaggedMetricsLocalSpanObserver(
                self.batch, span, self.allowlist, self.sample_rate
            )
        else:
            observer = TaggedMetricsClientSpanObserver(
                self.batch, span, self.allowlist, self.sample_rate
            )
        span.register(observer)

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        filtered_tags = {k: v for (k, v) in self.tags.items() if k in self.allowlist}

        for key, delta in self.counters.items():
            self.batch.counter(key, filtered_tags).increment(delta, sample_rate=self.sample_rate)

        self.timer.update_tags(filtered_tags)
        self.timer.stop()

        self.batch.counter(
            f"{self.base_name}.rate", {**filtered_tags, "success": not exc_info}
        ).increment(sample_rate=self.sample_rate)

        self.batch.flush()


class TaggedMetricsLocalSpanObserver(SpanObserver):
    def __init__(
        self, batch: metrics.Batch, span: Span, allowlist: Set[str], sample_rate: float = 1.0
    ):
        self.batch = batch
        self.span = span
        self.tags: Dict[str, Any] = {}
        self.base_name = "baseplate.local"
        self.timer = batch.timer(f"{self.base_name}.latency")
        self.allowlist = allowlist
        self.counters: Dict[str, float] = {}
        self.sample_rate = sample_rate

    def on_start(self) -> None:
        self.timer.start(self.sample_rate)
        self.tags["endpoint"] = self.span.name

    def on_incr_tag(self, key: str, delta: float) -> None:
        self.counters[key] = delta

    def on_set_tag(self, key: str, value: Any) -> None:
        self.tags[key] = value

    def on_child_span_created(self, span: Span) -> None:
        observer: SpanObserver
        if isinstance(span, LocalSpan):
            observer = TaggedMetricsLocalSpanObserver(
                self.batch, span, self.allowlist, self.sample_rate
            )
        else:
            observer = TaggedMetricsClientSpanObserver(
                self.batch, span, self.allowlist, self.sample_rate
            )
        span.register(observer)

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        filtered_tags = {k: v for (k, v) in self.tags.items() if k in self.allowlist}

        for key, delta in self.counters.items():
            self.batch.counter(key, filtered_tags).increment(delta, sample_rate=self.sample_rate)

        self.timer.update_tags(filtered_tags)
        self.timer.stop()

        self.batch.counter(
            f"{self.base_name}.rate",
            {**filtered_tags, "success": not exc_info},
        ).increment(sample_rate=self.sample_rate)

        self.batch.flush()


class TaggedMetricsClientSpanObserver(SpanObserver):
    def __init__(
        self, batch: metrics.Batch, span: Span, allowlist: Set[str], sample_rate: float = 1.0
    ):
        self.batch = batch
        self.span = span
        self.base_name = "baseplate.client"
        self.tags: Dict[str, Any] = {}
        self.timer = batch.timer(f"{self.base_name}.latency")
        self.allowlist = allowlist
        self.counters: Dict[str, float] = {}
        self.sample_rate = sample_rate

    def on_start(self) -> None:
        self.timer.start(self.sample_rate)
        self.tags["client"], _, self.tags["endpoint"] = self.span.name.rpartition(".")

    def on_incr_tag(self, key: str, delta: float) -> None:
        self.counters[key] = delta

    def on_set_tag(self, key: str, value: Any) -> None:
        self.tags[key] = value

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        filtered_tags = {k: v for (k, v) in self.tags.items() if k in self.allowlist}

        for key, delta in self.counters.items():
            self.batch.counter(key, filtered_tags).increment(delta, sample_rate=self.sample_rate)

        self.timer.update_tags(filtered_tags)
        self.timer.stop()

        self.batch.counter(
            f"{self.base_name}.rate",
            {**filtered_tags, "success": not exc_info},
        ).increment(sample_rate=self.sample_rate)

        self.batch.flush()



from __future__ import annotations

import time

from typing import Any
from typing import Dict
from typing import Optional

import pytest

from baseplate import RequestContext
from baseplate import ServerSpan
from baseplate import Span
from baseplate.lib.metrics import Counter
from baseplate.lib.metrics import Gauge
from baseplate.lib.metrics import Histogram
from baseplate.lib.metrics import Timer
from baseplate.observers.metrics_tagged import TaggedMetricsClientSpanObserver
from baseplate.observers.metrics_tagged import TaggedMetricsLocalSpanObserver
from baseplate.observers.metrics_tagged import TaggedMetricsServerSpanObserver


class TestException(Exception):
    pass


class FakeTimer:
    def __init__(self, batch: FakeBatch, name: str, tags: Dict[str, Any]):
        self.batch = batch
        self.name = name
        self.tags = tags

        self.start_time: Optional[float] = None
        self.sample_rate: float = 1.0

    def start(self, sample_rate: float = 1.0) -> None:
        self.start_time = time.time()
        self.sample_rate = sample_rate

    def stop(self) -> None:
        self.send(time.time() - self.start_time, self.sample_rate)

    def __enter__(self) -> None:
        self.start()

    def __exit__(self, *args) -> None:
        self.stop()
        return None

    def send(self, elapsed: float, sample_rate: float = 1.0) -> None:
        self.batch.timers.append(
            {"name": self.name, "elapsed": elapsed, "sample_rate": sample_rate, "tags": self.tags}
        )

    def update_tags(self, tags: Dict[str, Any]) -> None:
        self.tags.update(tags)


class FakeCounter:
    def __init__(self, batch: FakeBatch, name: str, tags: Dict[str, Any]):
        self.batch = batch
        self.name = name
        self.tags = tags

    def increment(self, delta: float = 1.0, sample_rate: float = 1.0) -> None:
        self.send(delta, sample_rate)

    def decrement(self, delta: float = 1.0, sample_rate: float = 1.0) -> None:
        self.increment(-delta, sample_rate)

    def send(self, delta: float, sample_rate: float) -> None:
        self.batch.counters.append(
            {"name": self.name, "delta": delta, "sample_rate": sample_rate, "tags": self.tags}
        )


class FakeBatch:
    def __init__(self):
        self.timers = []
        self.counters = []
        self.flushed = False

    def timer(self, name: str, tags: Optional[Dict[str, Any]] = None) -> Timer:
        return FakeTimer(self, name, tags or {})

    def counter(self, name: str, tags: Optional[Dict[str, Any]] = None) -> Counter:
        return FakeCounter(self, name, tags or {})

    def gauge(self, name: str, tags: Optional[Dict[str, Any]] = None) -> Gauge:
        raise NotImplementedError

    def histogram(self, name: str, tags: Optional[Dict[str, Any]] = None) -> Histogram:
        raise NotImplementedError

    def flush(self):
        self.flushed = True


@pytest.mark.parametrize(
    "observer_cls,name",
    (
        (TaggedMetricsServerSpanObserver, "server"),
        (TaggedMetricsClientSpanObserver, "client"),
        (TaggedMetricsLocalSpanObserver, "local"),
    ),
)
def test_observer(observer_cls, name):
    batch = FakeBatch()
    span = Span(
        trace_id=1234,
        parent_id=2345,
        span_id=3456,
        sampled=None,
        flags=None,
        name="fancy.span",
        context=RequestContext({}),
    )
    allow_list = {"client", "endpoint", "tag2"}
    sample_rate = 0.3
    observer = observer_cls(batch, span, allow_list, sample_rate=0.3)

    observer.on_start()
    observer.on_incr_tag("my.tag", 2)
    observer.on_set_tag("tag1", "foo")
    observer.on_set_tag("tag2", "bar")
    observer.on_finish(None)

    assert batch.timers
    timer = batch.timers.pop()
    assert batch.timers == []
    assert timer["sample_rate"] == sample_rate
    assert timer["name"] == f"baseplate.{name}.latency"

    if name in ("server", "local"):
        assert timer["tags"].pop("endpoint") == "fancy.span"
    else:
        assert timer["tags"].pop("client") == "fancy"
        assert timer["tags"].pop("endpoint") == "span"
    assert timer["tags"].pop("tag2") == "bar"
    assert timer["tags"] == {}

    for _ in range(2):
        counter = batch.counters.pop()
        assert counter["sample_rate"] == sample_rate
        if counter["name"] == f"baseplate.{name}.rate":
            assert counter["delta"] == 1
            assert counter["tags"].pop("success") is True
        elif counter["name"] == "my.tag":
            assert counter["delta"] == 2
        else:
            raise Exception(f"unexpected counter: {counter}")

        if name in ("server", "local"):
            assert counter["tags"].pop("endpoint") == "fancy.span"
        else:
            assert counter["tags"].pop("client") == "fancy"
            assert counter["tags"].pop("endpoint") == "span"
        assert counter["tags"].pop("tag2") == "bar"
        assert counter["tags"] == {}


def test_nested():
    batch = FakeBatch()
    span = ServerSpan(
        trace_id=1234,
        parent_id=2345,
        span_id=3456,
        sampled=None,
        flags=None,
        name="fancy.span",
        context=RequestContext({}),
    )
    observer = TaggedMetricsLocalSpanObserver(batch, span, {"client", "endpoint"})
    span.register(observer)

    with span.make_child("foo", local=True, component_name="foo") as child_span:
        assert len(child_span.observers) == 1



import json
import logging
import queue
import socket
import time

from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import TYPE_CHECKING

import confluent_kafka

from gevent.server import StreamServer
from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram

from baseplate import Baseplate
from baseplate import RequestContext
from baseplate.lib.prometheus_metrics import default_latency_buckets
from baseplate.server.queue_consumer import HealthcheckCallback
from baseplate.server.queue_consumer import make_simple_healthchecker
from baseplate.server.queue_consumer import MessageHandler
from baseplate.server.queue_consumer import PumpWorker
from baseplate.server.queue_consumer import QueueConsumerFactory


if TYPE_CHECKING:
    WorkQueue = queue.Queue[confluent_kafka.Message]  # pylint: disable=unsubscriptable-object
else:
    WorkQueue = queue.Queue


logger = logging.getLogger(__name__)


KafkaMessageDeserializer = Callable[[bytes], Any]
Handler = Callable[[RequestContext, Any, confluent_kafka.Message], None]


class KafkaConsumerPrometheusLabels(NamedTuple):
    kafka_client_name: str
    kafka_topic: str


KAFKA_PROCESSING_TIME = Histogram(
    "kafka_consumer_message_processing_time_seconds",
    "latency histogram of how long it takes to process a message",
    KafkaConsumerPrometheusLabels._fields + ("kafka_success",),
    buckets=default_latency_buckets,
)

KAFKA_PROCESSED_TOTAL = Counter(
    "kafka_consumer_messages_processed_total",
    "total count of messages processed by this host",
    KafkaConsumerPrometheusLabels._fields + ("kafka_success",),
)

KAFKA_ACTIVE_MESSAGES = Gauge(
    "kafka_consumer_active_messages",
    "gauge that reflects the number of messages currently being processed",
    KafkaConsumerPrometheusLabels._fields,
)


class KafkaConsumerWorker(PumpWorker):
    """Reads messages from the Kafka consumer and pumps them into the internal work_queue."""

    def __init__(
        self,
        baseplate: Baseplate,
        name: str,
        consumer: confluent_kafka.Consumer,
        work_queue: WorkQueue,
        batch_size: int = 1,
    ):
        self.baseplate = baseplate
        self.name = name
        self.consumer = consumer
        self.work_queue = work_queue
        self.batch_size = batch_size

        self.started = False
        self.stopped = False

    def run(self) -> None:
        logger.debug("Starting KafkaConsumerWorker.")
        self.started = True
        while not self.stopped:
            context = self.baseplate.make_context_object()
            with self.baseplate.make_server_span(context, f"{self.name}.pump") as span:
                with span.make_child("kafka.consume"):
                    messages = self.consumer.consume(num_messages=self.batch_size, timeout=0)

                if not messages:
                    logger.debug("waited 1s and received no messages, waiting again")
                    # we can't use timeouts on the kafka consumer methods
                    # because they call out to C code that ends up blocking
                    # the event loop. gevent will successfully yield control
                    # on this sleep call though.
                    # see https://tech.wayfair.com/2018/07/blocking-io-in-gunicorn-gevent-workers/
                    time.sleep(1)
                    continue

                with span.make_child("kafka.work_queue_put"):
                    for message in messages:
                        self.work_queue.put(message)

    def stop(self) -> None:
        # stop consuming, but leave the consumer instance intact. if we
        # close the consumer before the message handler is done it won't be able
        # to commit offsets
        logger.debug("Stopping KafkaConsumerWorker.")
        self.stopped = True


class KafkaMessageHandler(MessageHandler):
    """Reads messages from the internal work_queue and processes them."""

    def __init__(
        self,
        baseplate: Baseplate,
        name: str,
        handler_fn: Handler,
        message_unpack_fn: KafkaMessageDeserializer,
        on_success_fn: Optional[Handler] = None,
        prometheus_client_name: str = "",
    ):
        self.baseplate = baseplate
        self.name = name
        self.handler_fn = handler_fn
        self.message_unpack_fn = message_unpack_fn
        self.on_success_fn = on_success_fn
        self.prometheus_client_name = prometheus_client_name

    def handle(self, message: confluent_kafka.Message) -> None:
        prom_labels = KafkaConsumerPrometheusLabels(
            kafka_client_name=self.prometheus_client_name,
            kafka_topic=message.topic() if message.topic() is not None else "",
        )
        prom_success = "true"
        start_time = time.perf_counter()
        context = self.baseplate.make_context_object()
        try:
            # We place the call to ``baseplate.make_server_span`` inside the
            # try/except block because we still want Baseplate to see and
            # handle the error (publish it to error reporting)
            with self.baseplate.make_server_span(
                context, f"{self.name}.handler"
            ) as span, KAFKA_ACTIVE_MESSAGES.labels(**prom_labels._asdict()).track_inprogress():
                error = message.error()
                if error:
                    prom_success = "false"
                    # this isn't a real message, but is an error from Kafka
                    raise ValueError(f"KafkaError: {error.str()}")

                topic = message.topic()
                offset = message.offset()
                partition = message.partition()

                span.set_tag("kind", "consumer")
                span.set_tag("kafka.topic", topic)
                span.set_tag("kafka.key", message.key())
                span.set_tag("kafka.partition", partition)
                span.set_tag("kafka.offset", offset)
                span.set_tag("kafka.timestamp", message.timestamp())

                blob: bytes = message.value()

                try:
                    data = self.message_unpack_fn(blob)
                except Exception:
                    prom_success = "false"
                    logger.error("skipping invalid message")
                    context.span.incr_tag(f"{self.name}.{topic}.invalid_message")
                    return

                try:
                    ingest_timestamp_ms = data["endpoint_timestamp"]
                    now_ms = int(time.time() * 1000)
                    message_latency = (now_ms - ingest_timestamp_ms) / 1000
                except (KeyError, TypeError):
                    # we can't guarantee that all publishers populate this field
                    # v2 events publishers (event collectors) do, but future
                    # kafka publishers may not
                    message_latency = None

                self.handler_fn(context, data, message)

                if self.on_success_fn:
                    self.on_success_fn(context, data, message)

                if message_latency is not None:
                    context.metrics.timer(f"{self.name}.{topic}.latency").send(message_latency)

                context.metrics.gauge(f"{self.name}.{topic}.offset.{partition}").replace(offset)
        except Exception:
            prom_success = "false"
            # let this exception crash the server so we'll stop processing messages
            # and won't commit offsets. when the server restarts it will get
            # this message again and try to process it.
            logger.exception(
                "Unhandled error while trying to process a message, terminating the server"
            )
            raise
        finally:
            KAFKA_PROCESSING_TIME.labels(
                **prom_labels._asdict(), kafka_success=prom_success
            ).observe(time.perf_counter() - start_time)
            KAFKA_PROCESSED_TOTAL.labels(**prom_labels._asdict(), kafka_success=prom_success).inc()


class _BaseKafkaQueueConsumerFactory(QueueConsumerFactory):
    def __init__(
        self,
        name: str,
        baseplate: Baseplate,
        consumer: confluent_kafka.Consumer,
        handler_fn: Handler,
        kafka_consume_batch_size: int = 1,
        message_unpack_fn: KafkaMessageDeserializer = json.loads,
        health_check_fn: Optional[HealthcheckCallback] = None,
        prometheus_client_name: str = "",
    ):
        """`_BaseKafkaQueueConsumerFactory` constructor.

        :param name: A name for your consumer process. Must look like "kafka_consumer.{group_name}"
        :param baseplate: The Baseplate set up for your consumer.
        :param consumer: An instance of :py:class:`~confluent_kafka.Consumer`.
        :param handler_fn: A `baseplate.frameworks.queue_consumer.kafka.Handler`
            function that will process an individual message.
        :param kafka_consume_batch_size: The number of messages the `KafkaConsumerWorker`
            reads from Kafka in each batch. Defaults to 1.
        :param message_unpack_fn: A function that takes one argument, the `bytes` message body
            and returns the message in the format the handler expects. Defaults to `json.loads`.
        :param health_check_fn: A `baseplate.server.queue_consumer.HealthcheckCallback`
            function that can be used to customize your health check.
        :param prometheus_client_name: the service-provided name for the client to identify the
            backends for kafka. MUST be user specified, MAY be blank if not specified

        """
        self.name = name
        self.baseplate = baseplate
        self.consumer = consumer
        self.handler_fn = handler_fn
        self.kafka_consume_batch_size = kafka_consume_batch_size
        self.message_unpack_fn = message_unpack_fn
        self.health_check_fn = health_check_fn
        self.prometheus_client_name = prometheus_client_name

    @classmethod
    def new(
        cls,
        name: str,
        baseplate: Baseplate,
        bootstrap_servers: str,
        group_id: str,
        topics: Sequence[str],
        handler_fn: Handler,
        kafka_consume_batch_size: int = 1,
        message_unpack_fn: KafkaMessageDeserializer = json.loads,
        health_check_fn: Optional[HealthcheckCallback] = None,
        kafka_config: Optional[Dict[str, Any]] = None,
        prometheus_client_name: str = "",
    ) -> "_BaseKafkaQueueConsumerFactory":
        """Return a new `_BaseKafkaQueueConsumerFactory`.

        This method will create the :py:class:`~confluent_kafka.Consumer` for you and is
        appropriate to use in most cases where you just need a basic consumer
        with sensible defaults.

        This method will also enforce naming standards for the Kafka consumer group
        and the baseplate server span.

        :param name: A name for your consumer process. Must look like "kafka_consumer.{group_name}"
        :param baseplate: The Baseplate set up for your consumer.
        :param bootstrap_servers: A comma delimited string of kafka brokers.
        :param group_id: The kafka consumer group id. Must look like "{service_name}.{group_name}"
            to help prevent collisions between services.
        :param topics: An iterable of kafka topics to consume from.
        :param handler_fn: A `baseplate.frameworks.queue_consumer.kafka.Handler`
            function that will process an individual message.
        :param kafka_consume_batch_size: The number of messages the `KafkaConsumerWorker`
            reads from Kafka in each batch. Defaults to 1.
        :param message_unpack_fn: A function that takes one argument, the `bytes` message body
            and returns the message in the format the handler expects. Defaults to `json.loads`.
        :param health_check_fn: A `baseplate.server.queue_consumer.HealthcheckCallback`
            function that can be used to customize your health check.
        :param kafka_config: An optional config for confluent_kafka.Consumer

        """
        service_name, _, group_name = group_id.partition(".")
        assert service_name and group_name, "group_id must start with 'SERVICENAME.'"
        assert name == f"kafka_consumer.{group_name}"

        consumer = cls.make_kafka_consumer(bootstrap_servers, group_id, topics, kafka_config)

        return cls(
            name=name,
            baseplate=baseplate,
            consumer=consumer,
            handler_fn=handler_fn,
            kafka_consume_batch_size=kafka_consume_batch_size,
            message_unpack_fn=message_unpack_fn,
            health_check_fn=health_check_fn,
            prometheus_client_name=prometheus_client_name,
        )

    @classmethod
    def _consumer_config(cls) -> Dict[str, Any]:
        raise NotImplementedError

    @classmethod
    def make_kafka_consumer(
        cls,
        bootstrap_servers: str,
        group_id: str,
        topics: Sequence[str],
        kafka_config: Optional[Dict[str, Any]] = None,
    ) -> confluent_kafka.Consumer:
        consumer_config = {
            "bootstrap.servers": bootstrap_servers,
            "group.id": group_id,
            # reset the offset to the latest offset when no stored offset exists.
            # this means that when a new consumer group is created it will only
            # process new messages.
            "auto.offset.reset": "latest",
        }
        consumer_config.update(cls._consumer_config())
        if kafka_config:
            consumer_config.update(kafka_config)
        consumer = confluent_kafka.Consumer(consumer_config)

        try:
            # we can allow a blocking timeout here because there is only one
            # consumer for the entire server
            metadata = consumer.list_topics(timeout=10)
        except confluent_kafka.KafkaException:
            logger.error("failed getting metadata from %s, exiting.", bootstrap_servers)
            raise

        all_topics = set(metadata.topics.keys())
        for topic in topics:
            assert (
                topic in all_topics
            ), f"topic '{topic}' does not exist. maybe it's misspelled or on a different kafka cluster?"

        # pylint: disable=unused-argument
        def log_assign(
            consumer: confluent_kafka.Consumer, partitions: List[confluent_kafka.TopicPartition]
        ) -> None:
            for topic_partition in partitions:
                logger.info("assigned %s/%s", topic_partition.topic, topic_partition.partition)

        # pylint: disable=unused-argument
        def log_revoke(
            consumer: confluent_kafka.Consumer, partitions: List[confluent_kafka.TopicPartition]
        ) -> None:
            for topic_partition in partitions:
                logger.info("revoked %s/%s", topic_partition.topic, topic_partition.partition)

        consumer.subscribe(topics, on_assign=log_assign, on_revoke=log_revoke)
        return consumer

    def build_pump_worker(self, work_queue: WorkQueue) -> KafkaConsumerWorker:
        return KafkaConsumerWorker(
            baseplate=self.baseplate,
            name=self.name,
            consumer=self.consumer,
            work_queue=work_queue,
            batch_size=self.kafka_consume_batch_size,
        )

    def build_message_handler(self) -> KafkaMessageHandler:
        return KafkaMessageHandler(
            self.baseplate,
            self.name,
            self.handler_fn,
            self.message_unpack_fn,
            prometheus_client_name=self.prometheus_client_name,
        )

    def build_health_checker(self, listener: socket.socket) -> StreamServer:
        return make_simple_healthchecker(listener, callback=self.health_check_fn)


class InOrderConsumerFactory(_BaseKafkaQueueConsumerFactory):
    """Factory for running a :py:class:`~baseplate.server.queue_consumer.QueueConsumerServer` using Kafka.

    The `InOrderConsumerFactory` attempts to achieve in order, exactly once
    message processing.

    This will run a single `KafkaConsumerWorker` that reads messages from Kafka and
    puts them into an internal work queue. Then it will run a single `KafkaMessageHandler`
    that reads messages from the internal work queue, processes them with the
    `handler_fn`, and then commits each message's offset to Kafka.

    This one-at-a-time, in-order processing ensures that when a failure happens
    during processing we don't commit its offset (or the offset of any later
    messages) and that when the server restarts it will receive the failed
    message and attempt to process it again. Additionally, because each
    message's offset is committed immediately after processing we should
    never process a message more than once.

    For most cases where you just need a basic consumer with sensible defaults
    you can use `InOrderConsumerFactory.new`.

    If you need more control, you can create the :py:class:`~confluent_kafka.Consumer`
    yourself and use the constructor directly.

    """

    # we need to ensure that only a single message handler worker exists (max_concurrency = 1)
    # otherwise we could have out of order processing and mess up committing offsets to kafka!
    message_handler_count = 0

    @classmethod
    def _consumer_config(cls) -> Dict[str, Any]:
        return {
            # The consumer sends periodic heartbeats on a separate thread to
            # indicate its liveness to the broker. If no heartbeats are received by
            # the broker for a group member within the session timeout (because the
            # consumer and heartbeat thread have died), the broker
            # will remove the consumer from the group and trigger a rebalance.
            "heartbeat.interval.ms": 3000,
            "session.timeout.ms": 10000,
            # Maximum allowed time between calls to consume messages. If this
            # interval is exceeded the consumer is considered failed and the group
            # will rebalance in order to reassign the partitions to another consumer
            # group member.
            # Note: It is recommended to set enable.auto.offset.store=false for
            # long-time processing applications and then explicitly store offsets
            # after message processing, to make sure offsets are not auto-committed
            # prior to processing has finished.
            "max.poll.interval.ms": 300000,
            # disable offset autocommit, we'll manually commit.
            "enable.auto.commit": "false",
        }

    def build_message_handler(self) -> KafkaMessageHandler:
        assert self.message_handler_count == 0, "Can only run 1 message handler!"

        self.message_handler_count += 1

        # pylint: disable=unused-argument
        def commit_offset(
            context: RequestContext, data: Any, message: confluent_kafka.Message
        ) -> None:
            logger.debug(
                "committing topic %s partition %s offset %s",
                message.topic(),
                message.partition(),
                message.offset(),
            )
            with context.span.make_child("kafka.commit"):
                self.consumer.commit(message=message, asynchronous=False)

        return KafkaMessageHandler(
            self.baseplate,
            self.name,
            self.handler_fn,
            self.message_unpack_fn,
            # commit offset after each successful message handle()
            on_success_fn=commit_offset,
            prometheus_client_name=self.prometheus_client_name,
        )


class FastConsumerFactory(_BaseKafkaQueueConsumerFactory):
    """Factory for running a :py:class:`~baseplate.server.queue_consumer.QueueConsumerServer` using Kafka.

    The `FastConsumerFactory` prioritizes high throughput over exactly once
    message processing.

    This will run a single `KafkaConsumerWorker` that reads messages from Kafka and
    puts them into an internal work queue. Then it will run multiple `KafkaMessageHandler`s
    that read messages from the internal work queue, processes them with the
    `handler_fn`. The number of `KafkaMessageHandler` processes is controlled
    by the `max_concurrency` parameter in the `~baseplate.server.queue_consumer.QueueConsumerServer`
    configuration. Kafka partition offsets are automatically committed by the
    `confluent_kafka.Consumer` every 5 seconds, so any message that has been
    read by the `KafkaConsumerWorker` could be committed, regardless of whether
    it has been processed.

    This server should be able to achieve very high message processing throughput
    due to the multiple `KafkaMessageHandler` processes and less frequent, background
    partition offset commits. This does come at a price though: messages may be
    processed out of order, not at all, or multiple times. This is appropriate
    when processing throughput is important and it's acceptable to skip messages
    or process messages more than once (maybe there is ratelimiting in the
    handler or somewhere downstream).

    Messages processed out of order:
    Messages are added to the internal work queue in order, but one worker may
    finish processing a "later" message before another worker finishes
    processing an "earlier" message.

    Messages never processed:
    If the server crashes it may not have processed some messages that have already
    had their offsets automatically committed. When the server restarts it won't
    read those messages.

    Messages processed more than once:
    If the server crashes it may have processed some messages but not yet
    committed their offsets. When the server restarts it will reprocess those
    messages.

    For most cases where you just need a basic consumer with sensible defaults
    you can use `FastConsumerFactory.new`.

    If you need more control, you can create the :py:class:`~confluent_kafka.Consumer`
    yourself and use the constructor directly.

    """

    # pylint: disable=unused-argument
    @staticmethod
    def _commit_callback(
        err: confluent_kafka.KafkaError, topic_partition_list: List[confluent_kafka.TopicPartition]
    ) -> None:
        # called after automatic commits
        for topic_partition in topic_partition_list:
            topic = topic_partition.topic
            partition = topic_partition.partition
            offset = topic_partition.offset

            if topic_partition.error:
                logger.error(
                    "commit error topic %s partition %s offset %s", topic, partition, offset
                )
            elif offset == confluent_kafka.OFFSET_INVALID:
                # we receive offsets for all partitions. an offset value of
                # OFFSET_INVALID means that no offset was committed.
                pass
            else:
                logger.debug(
                    "commit success topic %s partition %s offset %s", topic, partition, offset
                )

    @classmethod
    def _consumer_config(cls) -> Dict[str, Any]:
        return {
            # The consumer sends periodic heartbeats on a separate thread to
            # indicate its liveness to the broker. If no heartbeats are received by
            # the broker for a group member within the session timeout (because the
            # consumer and heartbeat thread have died), the broker
            # will remove the consumer from the group and trigger a rebalance.
            "heartbeat.interval.ms": 3000,
            "session.timeout.ms": 10000,
            # Maximum allowed time between calls to consume messages. If this
            # interval is exceeded the consumer is considered failed and the group
            # will rebalance in order to reassign the partitions to another consumer
            # group member.
            # Note: It is recommended to set enable.auto.offset.store=false for
            # long-time processing applications and then explicitly store offsets
            # after message processing, to make sure offsets are not auto-committed
            # prior to processing has finished.
            "max.poll.interval.ms": 300000,
            # autocommit offsets every 5 seconds
            "enable.auto.commit": "true",
            "auto.commit.interval.ms": 5000,
            "enable.auto.offset.store": "true",
            # register a commit callback so that we'll log commits
            "on_commit": cls._commit_callback,
        }



from pylint.lint import PyLinter

from baseplate.lint.db_query_string_format_plugin import NoDbQueryStringFormatChecker


def register(linter: PyLinter) -> None:
    checker = NoDbQueryStringFormatChecker(linter)
    linter.register_checker(checker)



from unittest import mock

import pytest

from prometheus_client import REGISTRY

from baseplate.clients.kombu import _KombuProducer
from baseplate.clients.kombu import AMQP_PROCESSED_TOTAL
from baseplate.clients.kombu import AMQP_PROCESSING_TIME
from baseplate.clients.kombu import connection_from_config
from baseplate.clients.kombu import exchange_from_config
from baseplate.clients.kombu import KombuThriftSerializer
from baseplate.lib.config import ConfigurationError
from baseplate.testing.lib.secrets import FakeSecretsStore

from ... import does_not_raise
from ...integration.test_thrift.ttypes import ExampleStruct


def secrets():
    return FakeSecretsStore(
        {
            "secrets": {
                "secret/rabbitmq/account": {
                    "type": "credential",
                    "username": "spez",
                    "password": "hunter2",
                }
            },
        }
    )


@pytest.mark.parametrize(
    "app_config,kwargs,expectation,expected",
    [
        ({}, {}, pytest.raises(ConfigurationError), {}),
        ({"rabbitmq.virtual_host": "test"}, {}, pytest.raises(ConfigurationError), {}),
        ({"amqp.hostname": "amqp://rabbit.local:5672"}, {}, pytest.raises(ConfigurationError), {}),
        (
            {
                "rabbitmq.hostname": "amqp://rabbit.local:5672",
                "rabbitmq.credentials_secret": "secret/rabbitmq/account",
            },
            {},
            pytest.raises(ValueError),
            {},
        ),
        (
            {"rabbitmq.hostname": "amqp://rabbit.local:5672"},
            {},
            does_not_raise(),
            {"hostname": "rabbit.local", "userid": None, "password": None, "virtual_host": "/"},
        ),
        (
            {"rabbitmq.hostname": "amqp://rabbit.local:5672", "rabbitmq.virtual_host": "test"},
            {},
            does_not_raise(),
            {"hostname": "rabbit.local", "userid": None, "password": None, "virtual_host": "test"},
        ),
        (
            {"rabbitmq.hostname": "amqp://rabbit.local:5672", "rabbitmq.virtual_host": "test"},
            {"userid": "spez", "password": "hunter2"},
            does_not_raise(),
            {
                "hostname": "rabbit.local",
                "userid": "spez",
                "password": "hunter2",
                "virtual_host": "test",
            },
        ),
        (
            {
                "rabbitmq.hostname": "amqp://rabbit.local:5672",
                "rabbitmq.credentials_secret": "secret/rabbitmq/account",
            },
            {"secrets": secrets()},
            does_not_raise(),
            {
                "hostname": "rabbit.local",
                "userid": "spez",
                "password": "hunter2",
                "virtual_host": "/",
            },
        ),
    ],
)
def test_connection_from_config(app_config, kwargs, expectation, expected):
    with expectation:
        connection = connection_from_config(app_config, prefix="rabbitmq.", **kwargs)
    for attr, value in expected.items():
        assert getattr(connection, attr) == value


class TestKombuThriftSerializer:
    @pytest.fixture
    def serializer(self):
        return KombuThriftSerializer[ExampleStruct](ExampleStruct)

    @pytest.fixture
    def req(self):
        return ExampleStruct(string_field="foo", int_field=42)

    @pytest.fixture
    def req_bytes(self):
        return b"\x0b\x00\x01\x00\x00\x00\x03foo\n\x00\x02\x00\x00\x00\x00\x00\x00\x00*\x00"

    def test_name(self, serializer):
        assert serializer.name == "thrift-ExampleStruct"

    def test_serialize(self, serializer, req, req_bytes):
        serialized = serializer.serialize(req)
        assert isinstance(serialized, bytes)
        assert serialized == req_bytes

    def test_deserialize(self, serializer, req, req_bytes):
        request = serializer.deserialize(req_bytes)
        assert isinstance(request, ExampleStruct)
        assert request == req

    def test_serialize_errors(self, serializer):
        with pytest.raises(TypeError):
            serializer.serialize({"foo": "bar"})

    @pytest.mark.parametrize(
        "input,exc", [(None, EOFError), (b"foo", TypeError), ("foo", TypeError), (1, TypeError)]
    )
    def test_deserialize_errors(self, input, exc, serializer):
        with pytest.raises(exc):
            serializer.deserialize(input)


class Test_KombuProducer:
    def setup(self):
        AMQP_PROCESSING_TIME.clear()
        AMQP_PROCESSED_TOTAL.clear()

    @pytest.fixture
    def app_config(self):
        yield {
            "rabbitmq.hostname": "amqp://rabbit.local:5672",
            "rabbitmq.exchange_type": "topic",
            "rabbitmq.exchange_name": "test_name",
        }

    @pytest.fixture
    def prefix(self):
        yield "rabbitmq."

    @pytest.fixture
    def connection(self, app_config, prefix):
        yield connection_from_config(app_config, prefix=prefix)

    @pytest.fixture
    def exchange(self, app_config, prefix):
        yield exchange_from_config(app_config, prefix=prefix)

    @pytest.fixture
    def producer(self):
        yield mock.MagicMock()

    @pytest.fixture
    def producer_pool(self, producer):
        pp = mock.MagicMock()
        pp.acquire().__enter__.return_value = producer
        yield pp

    @pytest.fixture
    def producers(self, connection, producer_pool):
        p = mock.MagicMock()
        p.__getitem__.return_value = producer_pool
        yield p

    @pytest.fixture
    def span(self):
        yield mock.MagicMock()

    @pytest.fixture
    def kombu_producer(self, span, connection, exchange, producers):
        yield _KombuProducer("name", span, connection, exchange, producers)

    @pytest.fixture
    def expected_labels(self):
        yield {
            "amqp_address": "rabbit.local:5672",
            "amqp_virtual_host": "/",
            "amqp_exchange_type": "topic",
            "amqp_exchange_name": "test_name",
        }

    def test__on_success(self, kombu_producer, expected_labels):
        kombu_producer._on_success(1)
        expected_labels["amqp_success"] = "true"
        assert (
            REGISTRY.get_sample_value(
                f"{AMQP_PROCESSING_TIME._name}_bucket", {**expected_labels, **{"le": "+Inf"}}
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(f"{AMQP_PROCESSED_TOTAL._name}_total", expected_labels) == 1
        )

    def test__on_error(self, kombu_producer, expected_labels):
        kombu_producer._on_error(1)
        expected_labels["amqp_success"] = "false"
        assert (
            REGISTRY.get_sample_value(
                f"{AMQP_PROCESSING_TIME._name}_bucket", {**expected_labels, **{"le": "+Inf"}}
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(f"{AMQP_PROCESSED_TOTAL._name}_total", expected_labels) == 1
        )

    def test_publish_prom_exc(self, kombu_producer, expected_labels, producer):
        producer.publish.side_effect = Exception("Any error")

        with pytest.raises(Exception, match=r"^Any error$"):
            kombu_producer.publish(body="doesn't matter")
        expected_labels["amqp_success"] = "false"
        assert (
            REGISTRY.get_sample_value(
                f"{AMQP_PROCESSING_TIME._name}_bucket", {**expected_labels, **{"le": "+Inf"}}
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(f"{AMQP_PROCESSED_TOTAL._name}_total", expected_labels) == 1
        )



from contextlib import nullcontext as does_not_raise
from unittest import mock

import pytest

from prometheus_client import REGISTRY
from requests import Request
from requests import Response
from requests import Session

from baseplate.clients.requests import ACTIVE_REQUESTS
from baseplate.clients.requests import BaseplateSession
from baseplate.clients.requests import LATENCY_SECONDS
from baseplate.clients.requests import REQUESTS_TOTAL
from baseplate.lib.prometheus_metrics import getHTTPSuccessLabel


@pytest.fixture
def baseplate_session(request):
    yield BaseplateSession(
        adapter=mock.MagicMock(),
        name="session_name",
        span=mock.MagicMock(),
        client_name=request.param,
    )


class TestBaseplateSessionProm:
    def setup(self):
        ACTIVE_REQUESTS.clear()
        LATENCY_SECONDS.clear()
        REQUESTS_TOTAL.clear()

    # request is a reserved name that can't be used in parametrize
    @pytest.mark.parametrize(
        "req",
        [
            Request("DELETE", "http://example.com/foo/bar"),
            Request("GET", "http://example.com/foo/bar"),
            Request("HEAD", "http://example.com/foo/bar"),
            Request("OPTIONS", "http://example.com/foo/bar"),
            Request("PATCH", "http://example.com/foo/bar"),
            Request("POST", "http://example.com/foo/bar"),
            Request("PUT", "http://example.com/foo/bar"),
        ],
    )
    @pytest.mark.parametrize(
        "response",
        [
            mock.MagicMock(spec=Response, status_code=200),
            mock.MagicMock(spec=Response, status_code=404),
            mock.MagicMock(spec=Response, status_code=503),
            Exception("smth went wrong"),
        ],
    )
    @pytest.mark.parametrize(
        "baseplate_session",
        [
            "test_client_name",
            "",
            None,
        ],
        indirect=True,
    )
    def test_send(self, req, response, baseplate_session):
        prom_labels = {
            "http_method": req.method.lower(),
            "http_client_name": baseplate_session.client_name
            if baseplate_session.client_name is not None
            else baseplate_session.name,
        }

        http_success = "true"
        status_code = ""
        expectation = does_not_raise()
        if isinstance(response, Exception):
            # if an exception is raised, success is false and status code is empty
            http_success = "false"
            expectation = pytest.raises(type(response))
        else:
            # otherwise we just grab that info from the response itself
            status_code = response.status_code
            http_success = getHTTPSuccessLabel(status_code)

        with mock.patch("baseplate.clients.requests.Session", spec=Session) as session:
            session().send.return_value = response
            with expectation:
                baseplate_session.send(req.prepare())

        assert REGISTRY.get_sample_value("http_client_active_requests", prom_labels) == 0

        assert (
            REGISTRY.get_sample_value(
                "http_client_latency_seconds_bucket",
                {**prom_labels, "http_success": http_success, "le": "+Inf"},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(
                "http_client_requests_total",
                {
                    **prom_labels,
                    "http_response_code": str(status_code),
                    "http_success": http_success,
                },
            )
            == 1
        )



import importlib
import queue
import unittest

import gevent.monkey
import gevent.queue

from baseplate.server.monkey import gevent_is_patched
from baseplate.server.monkey import patch_stdlib_queues


class MonkeyPatchTests(unittest.TestCase):
    def tearDown(self):
        importlib.reload(queue)
        gevent.monkey.saved.clear()

    def test_patch_stdlib_queues(self):
        assert queue.LifoQueue is not gevent.queue.LifoQueue
        patch_stdlib_queues()
        assert queue.LifoQueue is gevent.queue.LifoQueue

    def test_is_gevent_patched(self):
        assert not gevent_is_patched()
        patch_stdlib_queues()
        assert gevent_is_patched()



import io
import socket
import sys
import unittest

from unittest import mock

import pytest

from baseplate import server
from baseplate.lib import config


EXAMPLE_ENDPOINT = config.EndpointConfiguration(socket.AF_INET, ("127.0.0.1", 1234))


class ParseArgsTests(unittest.TestCase):
    def test_no_args(self):
        with mock.patch("sys.stderr", mock.Mock()):
            with self.assertRaises(SystemExit):
                server.parse_args([])

    def test_filename(self):
        with mock.patch("argparse.FileType", autospec=True) as make_file:
            args = server.parse_args(["filename"])

        mock_file = make_file.return_value
        self.assertEqual(mock_file.call_args, mock.call("filename"))
        self.assertEqual(args.config_file, mock_file.return_value)

    @mock.patch("argparse.FileType", autospec=True)
    def test_options(self, make_file):
        args = server.parse_args(
            [
                "filename",
                "--debug",
                "--app-name",
                "app",
                "--server-name",
                "server",
                "--bind",
                "1.2.3.4:81",
            ]
        )
        self.assertTrue(args.debug)
        self.assertEqual(args.app_name, "app")
        self.assertEqual(args.server_name, "server")
        self.assertEqual(args.bind, config.EndpointConfiguration(socket.AF_INET, ("1.2.3.4", 81)))


class MakeListenerTests(unittest.TestCase):
    @mock.patch("baseplate.server.einhorn.get_socket")
    @mock.patch("baseplate.server.einhorn.is_worker")
    def test_einhorn_managed(self, is_worker, get_socket):
        is_worker.return_value = True

        listener = server.make_listener(EXAMPLE_ENDPOINT)

        self.assertEqual(listener, get_socket.return_value)

    @mock.patch.dict("os.environ", {}, clear=True)
    @mock.patch("fcntl.fcntl")
    @mock.patch("socket.socket")
    def test_manually_bound(self, mocket, fcntl):
        listener = server.make_listener(EXAMPLE_ENDPOINT)

        self.assertEqual(mocket.call_args, mock.call(socket.AF_INET, socket.SOCK_STREAM))
        self.assertEqual(listener, mocket.return_value)

        self.assertEqual(listener.bind.call_args, mock.call(("127.0.0.1", 1234)))


class LoadFactoryTests(unittest.TestCase):
    @mock.patch("importlib.import_module", autospec=True)
    def test_full_url(self, import_module):
        factory = server._load_factory("package.module:callable", "default_name")

        self.assertEqual(import_module.call_args, mock.call("package.module"))
        self.assertEqual(factory, import_module.return_value.callable)

    @mock.patch("importlib.import_module", autospec=True)
    def test_default_name(self, import_module):
        factory = server._load_factory("package.module", "default_name")

        self.assertEqual(import_module.call_args, mock.call("package.module"))
        self.assertEqual(factory, import_module.return_value.default_name)


class CheckFnSignatureTests(unittest.TestCase):
    def test_no_args(self):
        def foo():
            pass

        with self.assertRaises(ValueError):
            server._fn_accepts_additional_args(foo, [])

    def test_var_args(self):
        def foo(*args):
            pass

        server._fn_accepts_additional_args(foo, [])
        server._fn_accepts_additional_args(foo, ["arg1"])
        server._fn_accepts_additional_args(foo, ["arg1", "arg2"])

    def test_config_arg_only(self):
        def foo(app_config):
            pass

        server._fn_accepts_additional_args(foo, [])
        with self.assertRaises(ValueError):
            server._fn_accepts_additional_args(foo, ["extra_arg"])

    def test_config_arg_with_var_args(self):
        def foo(app_config, *args):
            pass

        server._fn_accepts_additional_args(foo, [])
        server._fn_accepts_additional_args(foo, ["arg1"])
        server._fn_accepts_additional_args(foo, ["arg1", "arg2"])

    def test_additional_args(self):
        def foo(app_config, args):
            pass

        server._fn_accepts_additional_args(foo, [])
        server._fn_accepts_additional_args(foo, ["arg1"])
        server._fn_accepts_additional_args(foo, ["arg1", "arg2"])

    def test_additional_args_with_var_args(self):
        def foo(app_config, args, *extra):
            pass

        server._fn_accepts_additional_args(foo, [])
        server._fn_accepts_additional_args(foo, ["arg1"])
        server._fn_accepts_additional_args(foo, ["arg1", "arg2"])

    def test_kwargs(self):
        def foo(app_config, arg1, *, bar, **kwargs):
            pass

        server._fn_accepts_additional_args(foo, [])
        server._fn_accepts_additional_args(foo, ["arg1", "arg2", "arg3"])
        server._fn_accepts_additional_args(foo, ["arg1"])


class ParseBaseplateScriptArgs(unittest.TestCase):
    @mock.patch.object(sys, "argv", ["baseplate-script", "mock.ini", "package.module:callable"])
    @mock.patch("baseplate.server._load_factory")
    @mock.patch("builtins.open", mock.mock_open())
    def test_simple_call(self, _load_factory):
        args, extra_args = server._parse_baseplate_script_args()
        self.assertEqual(args.app_name, "main")
        self.assertEqual(extra_args, [])

    @mock.patch.object(
        sys, "argv", ["baseplate-script", "mock.ini", "package.module:callable", "--app-name", "ci"]
    )
    @mock.patch("baseplate.server._load_factory")
    @mock.patch("builtins.open", mock.mock_open())
    def test_specifying_app_name(self, _load_factory):
        args, extra_args = server._parse_baseplate_script_args()
        self.assertEqual(args.app_name, "ci")
        self.assertEqual(extra_args, [])

    @mock.patch.object(
        sys,
        "argv",
        ["baseplate-script", "mock.ini", "package.module:callable", "extra_arg1", "extra_arg2"],
    )
    @mock.patch("baseplate.server._load_factory")
    @mock.patch("builtins.open", mock.mock_open())
    def test_extra_args(self, _load_factory):
        args, extra_args = server._parse_baseplate_script_args()
        self.assertEqual(args.app_name, "main")
        self.assertEqual(extra_args, ["extra_arg1", "extra_arg2"])


@mock.patch.dict("os.environ", {"FOO_FROM_ENV": "environmental"})
@pytest.mark.parametrize(
    "config_text,expected",
    (
        ("", None),
        ("foo = bar", "bar"),
        ("foo = $FOO_FROM_ENV", "environmental"),
        ("foo = ${FOO_FROM_ENV}", "environmental"),
        ("foo = ${this:is:not:valid}", "${this:is:not:valid}"),
    ),
)
def test_read_config(config_text, expected):
    config_file = io.StringIO(f"[app:main]\n{config_text}\n")
    config_file.name = "<test>"
    config = server.read_config(config_file, server_name=None, app_name="main")
    assert config.app.get("foo") == expected



import socket
import unittest

from unittest import mock

from baseplate.server import einhorn


class NotEinhornWorkerTests(unittest.TestCase):
    def test_is_not_worker(self):
        self.assertFalse(einhorn.is_worker())

    def test_get_socket_count(self):
        with self.assertRaises(einhorn.NotEinhornWorker):
            einhorn.get_socket_count()

    def test_get_socket(self):
        with self.assertRaises(einhorn.NotEinhornWorker):
            einhorn.get_socket()

    def test_send_ack(self):
        with self.assertRaises(einhorn.NotEinhornWorker):
            einhorn.ack_startup()


@mock.patch.dict(
    "os.environ",
    {
        "EINHORN_MASTER_PID": "123",
        "EINHORN_SOCK_PATH": "/tmp/einhorn.sock",
        "EINHORN_FD_COUNT": "2",
        "EINHORN_FD_0": "5",
        "EINHORN_FD_1": "6",
        "EINHORN_FD_FAMILY_1": "AF_UNIX",
    },
    clear=True,
)
class EinhornWorkerTests(unittest.TestCase):
    def setUp(self):
        getppid_patcher = mock.patch("os.getppid")
        getppid = getppid_patcher.start()
        getppid.return_value = 123
        self.addCleanup(getppid_patcher.stop)

    def test_is_worker(self):
        self.assertTrue(einhorn.is_worker())

    def test_get_socket_count(self):
        self.assertEqual(einhorn.get_socket_count(), 2)

    @mock.patch("socket.fromfd", autospec=True)
    def test_get_socket(self, fromfd):
        sock = einhorn.get_socket()
        fromfd.assert_called_with(5, socket.AF_INET, socket.SOCK_STREAM)
        self.assertEqual(sock, fromfd.return_value)

    @mock.patch("socket.fromfd", autospec=True)
    def test_get_socket_unix(self, fromfd):
        sock = einhorn.get_socket(1)
        fromfd.assert_called_with(6, socket.AF_UNIX, socket.SOCK_STREAM)
        self.assertEqual(sock, fromfd.return_value)

    def test_get_socket_out_of_bounds(self):
        with self.assertRaises(IndexError):
            einhorn.get_socket(-1)

        with self.assertRaises(IndexError):
            einhorn.get_socket(2)

    @mock.patch("socket.socket")
    @mock.patch("os.getpid", autospec=True)
    def test_send_ack(self, getpid, sock):
        getpid.return_value = 1337

        einhorn.ack_startup()

        sock.assert_called_with(socket.AF_UNIX, socket.SOCK_STREAM)
        mocket = sock.return_value
        mocket.connect.assert_called_with("/tmp/einhorn.sock")

        self.assertEqual(mocket.sendall.call_count, 1)
        self.assertEqual(mocket.close.call_count, 1)



from gevent import monkey


def patch_stdlib_queues() -> None:
    """It is a common pattern to use a queue as a connection pool.
    This pattern can leak connections if the queue is not patched by gevent
    Gevent doesn't patch most queues by default.
    https://github.com/gevent/gevent/issues/1875
    """
    import queue
    import gevent.queue

    monkey.patch_module(queue, gevent.queue, items=["Queue", "LifoQueue", "PriorityQueue"])


def gevent_is_patched() -> bool:
    """Returns true if gevent is patched, otherwise returns false."""
    return bool(monkey.saved)



import unittest

from time import sleep
from uuid import uuid4

from baseplate import Baseplate
from baseplate.lib.ratelimit import RateLimiterContextFactory
from baseplate.lib.ratelimit import RateLimitExceededException

try:
    from pymemcache.client.base import PooledClient
    from baseplate.lib.ratelimit.backends.memcache import MemcacheRateLimitBackendContextFactory
except ImportError:
    raise unittest.SkipTest("pymemcache is not installed")
try:
    from redis import ConnectionPool
    from baseplate.lib.ratelimit.backends.redis import RedisRateLimitBackendContextFactory
except ImportError:
    raise unittest.SkipTest("redis-py is not installed")

from . import TestBaseplateObserver, get_endpoint_or_skip_container


redis_endpoint = get_endpoint_or_skip_container("redis", 6379)
memcached_endpoint = get_endpoint_or_skip_container("memcached", 11211)


class RateLimiterBackendTests:
    def setUp(self):
        self.allowance = 10
        self.interval = 1
        ratelimiter_factory = RateLimiterContextFactory(
            self.backend_factory, self.allowance, self.interval
        )

        self.baseplate_observer = TestBaseplateObserver()

        baseplate = Baseplate()
        baseplate.register(self.baseplate_observer)
        baseplate.add_to_context("ratelimiter", ratelimiter_factory)

        self.context = baseplate.make_context_object()
        self.server_span = baseplate.make_server_span(self.context, "test")

    def test_ratelimiter_consume(self):
        user_id = str(uuid4())
        with self.server_span:
            self.context.ratelimiter.consume(user_id, amount=self.allowance)

    def test_ratelimiter_exceeded(self):
        user_id = str(uuid4())
        with self.server_span:
            with self.assertRaises(RateLimitExceededException):
                self.context.ratelimiter.consume(user_id, amount=self.allowance + 1)

    def test_ratelimiter_resets(self):
        user_id = str(uuid4())
        with self.server_span:
            self.context.ratelimiter.consume(user_id, amount=self.allowance)
            sleep(self.interval)
            self.context.ratelimiter.consume(user_id, amount=self.allowance)


class RedisRateLimitBackendTests(RateLimiterBackendTests, unittest.TestCase):
    def setUp(self):
        pool = ConnectionPool(host=redis_endpoint.address.host, port=redis_endpoint.address.port)
        self.backend_factory = RedisRateLimitBackendContextFactory(pool)
        super().setUp()


class MemcacheRateLimitBackendTests(RateLimiterBackendTests, unittest.TestCase):
    def setUp(self):
        pool = PooledClient(server=memcached_endpoint.address)
        self.backend_factory = MemcacheRateLimitBackendContextFactory(pool)
        super().setUp()



import grp
import json
import os
import pwd
import tempfile
import unittest

from pathlib import Path

import boto3

from moto import mock_s3

from baseplate.sidecars.live_data_watcher import NodeWatcher


class NodeWatcherTests(unittest.TestCase):
    mock_s3 = mock_s3()

    def setUp(self):
        self.mock_s3.start()
        bucket_name = "test_bucket"
        s3_data = {"foo_encrypted": "bar_encrypted"}
        s3_client = boto3.client(
            "s3",
            region_name="us-east-1",
        )
        s3_client.create_bucket(Bucket=bucket_name)
        s3_client.put_object(
            Bucket=bucket_name,
            Key="test_file_key",
            Body=json.dumps(s3_data).encode(),
            SSECustomerKey="test_decryption_key",
            SSECustomerAlgorithm="AES256",
        )

    def tearDown(self):
        self.mock_s3.stop()

    def run(self, result: unittest.TestResult = None) -> unittest.TestResult:
        with tempfile.TemporaryDirectory(prefix=self.id()) as loc:
            self.output_dir = Path(loc)
            return super().run(result)

    def test_s3_load_type_on_change(self):
        dest = self.output_dir.joinpath("data.txt")
        inst = NodeWatcher(str(dest), os.getuid(), os.getgid(), 777)

        new_content = b'{"live_data_watcher_load_type":"S3","bucket_name":"test_bucket","file_key":"test_file_key","sse_key":"test_decryption_key","region_name":"us-east-1"}'
        expected_content = b'{"foo_encrypted": "bar_encrypted"}'
        inst.on_change(new_content, None)
        self.assertEqual(expected_content, dest.read_bytes())
        self.assertEqual(dest.owner(), pwd.getpwuid(os.getuid()).pw_name)
        self.assertEqual(dest.group(), grp.getgrgid(os.getgid()).gr_name)

    def test_on_change(self):
        dest = self.output_dir.joinpath("data.txt")
        inst = NodeWatcher(str(dest), os.getuid(), os.getgid(), 777)

        new_content = b"foobar"
        inst.on_change(new_content, None)
        self.assertEqual(new_content, dest.read_bytes())
        self.assertEqual(dest.owner(), pwd.getpwuid(os.getuid()).pw_name)
        self.assertEqual(dest.group(), grp.getgrgid(os.getgid()).gr_name)

    def test_on_change_no_data(self):
        dest = self.output_dir.joinpath("data.txt")
        inst = NodeWatcher(str(dest), os.getuid(), os.getgid(), 777)

        new_content = None
        inst.on_change(new_content, None)
        self.assertEqual(False, os.path.exists(dest))

    def test_on_change_new_dir(self):
        dest = self.output_dir.joinpath("data/output.json")
        inst = NodeWatcher(str(dest), os.getuid(), os.getgid(), 777)

        new_content = b"foobar"
        inst.on_change(new_content, None)
        self.assertEqual(new_content, dest.read_bytes())

    def test_on_change_deep_new_dir(self):
        dest = self.output_dir.joinpath("data/foo/bar.json")
        inst = NodeWatcher(str(dest), os.getuid(), os.getgid(), 777)

        new_content = b"foobar"
        inst.on_change(new_content, None)
        self.assertEqual(new_content, dest.read_bytes())



from wsgiref.simple_server import make_server

from pyramid.config import Configurator
from pyramid.view import view_config


@view_config(route_name="hello_world", renderer="json")
def hello_world(request):
    return {"Hello": "World"}


def make_wsgi_app():
    configurator = Configurator()
    configurator.add_route("hello_world", "/", request_method="GET")
    configurator.scan()
    return configurator.make_wsgi_app()


if __name__ == "__main__":
    app = make_wsgi_app()
    server = make_server("127.0.0.1", 9090, app)
    server.serve_forever()



import abc
import time

from typing import Any
from typing import Generic
from typing import Optional
from typing import Type
from typing import TypeVar

import kombu.serialization

from kombu import Connection
from kombu import Exchange
from kombu.pools import Producers
from prometheus_client import Counter
from prometheus_client import Histogram
from thrift import TSerialization
from thrift.protocol.TBinaryProtocol import TBinaryProtocolAcceleratedFactory
from thrift.protocol.TProtocol import TProtocolFactory
from vine.promises import promise  # type: ignore

from baseplate import Span
from baseplate.clients import ContextFactory
from baseplate.lib import config
from baseplate.lib.prometheus_metrics import default_latency_buckets
from baseplate.lib.secrets import SecretsStore


T = TypeVar("T")

amqp_producer_labels = [
    "amqp_address",
    "amqp_virtual_host",
    "amqp_exchange_type",
    "amqp_exchange_name",
    "amqp_success",
]

AMQP_PROCESSING_TIME = Histogram(
    "amqp_producer_message_processing_seconds",
    "latency histogram of how long it takes to queue a message",
    amqp_producer_labels,
    buckets=default_latency_buckets,
)
AMQP_PROCESSED_TOTAL = Counter(
    "amqp_producer_message_processed_total",
    "total messages produced by this host",
    amqp_producer_labels,
)


def connection_from_config(
    app_config: config.RawConfig, prefix: str, secrets: Optional[SecretsStore] = None, **kwargs: Any
) -> Connection:
    """Make a Connection from a configuration dictionary.

    The keys useful to :py:func:`connection_from_config` should be prefixed,
    e.g. ``amqp.hostname`` etc. The ``prefix`` argument specifies the
    prefix used to filter keys.  Each key is mapped to a corresponding keyword
    argument on the :py:class:`~kombu.connection.Connection` constructor.  Any
    keyword arguments given to this function will be passed through to the
    :py:class:`~kombu.connection.Connection` constructor. Keyword arguments
    take precedence over the configuration file.

    Supported keys:

    * ``credentials_secret``
    * ``hostname``
    * ``virtual_host``

    """
    assert prefix.endswith(".")
    parser = config.SpecParser(
        {
            "credentials_secret": config.Optional(config.String),
            "hostname": config.String,
            "virtual_host": config.Optional(config.String),
        }
    )
    options = parser.parse(prefix[:-1], app_config)
    if options.credentials_secret:
        if not secrets:
            raise ValueError("'secrets' is required if 'credentials_secret' is set")
        credentials = secrets.get_credentials(options.credentials_secret)
        kwargs.setdefault("userid", credentials.username)
        kwargs.setdefault("password", credentials.password)
    return Connection(hostname=options.hostname, virtual_host=options.virtual_host, **kwargs)


def exchange_from_config(app_config: config.RawConfig, prefix: str, **kwargs: Any) -> Exchange:
    """Make an Exchange from a configuration dictionary.

    The keys useful to :py:func:`exchange_from_config` should be prefixed,
    e.g. ``amqp.exchange_name`` etc. The ``prefix`` argument specifies the
    prefix used to filter keys.  Each key is mapped to a corresponding keyword
    argument on the :py:class:`~kombu.Exchange` constructor.  Any keyword
    arguments given to this function will be passed through to the
    :py:class:`~kombu.Exchange` constructor. Keyword arguments take precedence
    over the configuration file.

    Supported keys:

    * ``exchange_name``
    * ``exchange_type``

    """
    assert prefix.endswith(".")
    parser = config.SpecParser(
        {"exchange_name": config.Optional(config.String), "exchange_type": config.String}
    )
    options = parser.parse(prefix[:-1], app_config)
    return Exchange(name=options.exchange_name or "", type=options.exchange_type, **kwargs)


class KombuSerializer(abc.ABC, Generic[T]):
    """Interface for wrapping non-built-in serializers for Kombu."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Return a unique name for the Serializer.

        This is used as the identifier for the serializer within Kombu so it
        should be unique.  It is also included in the "content-type" header so
        consumers are able to identify which serializer to use (in the format
        "application/x-{name}").
        """

    @abc.abstractmethod
    def serialize(self, obj: T) -> bytes:
        """Serialize the object into bytes for publishing."""

    @abc.abstractmethod
    def deserialize(self, message: bytes) -> T:
        """Deserialize the message bytes into an object for consuming."""


class KombuThriftSerializer(KombuSerializer[T]):  # pylint: disable=unsubscriptable-object
    """Thrift object serializer for Kombu."""

    def __init__(
        self,
        thrift_class: Type[T],
        protocol_factory: TProtocolFactory = TBinaryProtocolAcceleratedFactory(),
    ):
        self.thrift_class = thrift_class
        self.factory = protocol_factory

    @property
    def name(self) -> str:
        return f"thrift-{self.thrift_class.__name__}"

    def serialize(self, obj: T) -> bytes:
        if not isinstance(obj, self.thrift_class):
            raise TypeError(f"object to serialize must be of {self.thrift_class.__name__} type")
        return TSerialization.serialize(obj, self.factory)

    def deserialize(self, message: bytes) -> T:
        return TSerialization.deserialize(self.thrift_class(), message, self.factory)


def register_serializer(serializer: KombuSerializer) -> None:
    """Register `serializer` with the Kombu serialization registry.

    The serializer will be registered using `serializer.name` and will be sent
    to the message broker with the header "application/x-{serializer.name}".
    You need to call `register_serializer` before you can use a serializer for
    automatic serialization when publishing and deserializing when consuming.
    """
    kombu.serialization.register(
        serializer.name,
        serializer.serialize,
        serializer.deserialize,
        content_type=f"application/x-{serializer.name}",
        content_encoding="binary",
    )


class KombuProducer(config.Parser):
    """Configure a Kombu producer.

    This is meant to be used with
    :py:meth:`baseplate.Baseplate.configure_context`.

    See :py:func:`connection_from_config` and :py:func:`exchange_from_config`
    for available configuration settings.

    :param max_connections: The maximum number of connections.
    :param serializer: A custom message serializer.
    :param secrets: `SecretsStore` for non-default connection credentials.
    """

    def __init__(
        self,
        max_connections: Optional[int] = None,
        serializer: Optional[KombuSerializer] = None,
        secrets: Optional[SecretsStore] = None,
    ):
        self.max_connections = max_connections
        self.serializer = serializer
        self.secrets = secrets

    def parse(self, key_path: str, raw_config: config.RawConfig) -> "KombuProducerContextFactory":
        connection = connection_from_config(raw_config, prefix=f"{key_path}.", secrets=self.secrets)
        exchange = exchange_from_config(raw_config, prefix=f"{key_path}.")
        return KombuProducerContextFactory(
            connection, exchange, max_connections=self.max_connections, serializer=self.serializer
        )


class KombuProducerContextFactory(ContextFactory):
    """KombuProducer context factory.

    This factory will attach a proxy object which acts like a
    :py:class:`kombu.Producer` to an attribute on the
    :py:class:`~baseplate.RequestContext`.  The
    :py:meth:`~baseplate.clients.kombu.KombuProducer.publish` method will
    automatically record diagnostic information.

    :param connection: A configured connection object.
    :param exchange: A configured exchange object
    :param max_connections: The maximum number of connections.

    """

    def __init__(
        self,
        connection: Connection,
        exchange: Exchange,
        max_connections: Optional[int] = None,
        serializer: Optional[KombuSerializer] = None,
    ):
        self.connection = connection
        self.exchange = exchange
        self.producers = Producers(limit=max_connections)
        self.serializer = serializer

    def make_object_for_context(self, name: str, span: Span) -> "_KombuProducer":
        return _KombuProducer(
            name, span, self.connection, self.exchange, self.producers, serializer=self.serializer
        )


class _KombuProducer:
    def __init__(
        self,
        name: str,
        span: Span,
        connection: Connection,
        exchange: Exchange,
        producers: Producers,
        serializer: Optional[KombuSerializer] = None,
    ):
        self.name = name
        self.span = span
        self.connection = connection
        self.exchange = exchange
        self.producers = producers
        self.serializer = serializer
        self.prom_labels = {
            "amqp_address": f"{self.connection.hostname}:{self.connection.port}",
            "amqp_virtual_host": self.connection.virtual_host,
            "amqp_exchange_type": self.exchange.type,
            "amqp_exchange_name": self.exchange.name,
        }

    def _on_success(self, start_time: float) -> None:
        if isinstance(self.connection.transport, kombu.transport.pyamqp.Transport):
            AMQP_PROCESSING_TIME.labels(**self.prom_labels, amqp_success="true").observe(
                time.perf_counter() - start_time
            )
            AMQP_PROCESSED_TOTAL.labels(**self.prom_labels, amqp_success="true").inc()

    def _on_error(self, start_time: float) -> None:
        if isinstance(self.connection.transport, kombu.transport.pyamqp.Transport):
            AMQP_PROCESSING_TIME.labels(**self.prom_labels, amqp_success="false").observe(
                time.perf_counter() - start_time
            )
            AMQP_PROCESSED_TOTAL.labels(**self.prom_labels, amqp_success="false").inc()

    def publish(self, body: Any, routing_key: Optional[str] = None, **kwargs: Any) -> Any:
        start_time = time.perf_counter()
        if self.serializer:
            kwargs.setdefault("serializer", self.serializer.name)

        trace_name = f"{self.name}.publish"
        child_span = self.span.make_child(trace_name)

        child_span.set_tag("kind", "producer")
        if routing_key:
            child_span.set_tag("message_bus.destination", routing_key)

        with child_span:
            producer_pool = self.producers[self.connection]
            with producer_pool.acquire(block=True) as producer:
                try:
                    p = producer.publish(
                        body=body, routing_key=routing_key, exchange=self.exchange, **kwargs
                    )
                    # since publish(...) returns a promise, we have to handle it with callbacks
                    on_success = promise(self._on_success, (start_time,))
                    on_error = promise(self._on_error, (start_time,))
                    p.then(on_success, on_error=on_error)
                    return p
                except Exception:
                    if isinstance(self.connection.transport, kombu.transport.pyamqp.Transport):
                        # we have to handle exceptions here
                        AMQP_PROCESSING_TIME.labels(
                            **self.prom_labels, amqp_success="false"
                        ).observe(time.perf_counter() - start_time)
                        AMQP_PROCESSED_TOTAL.labels(**self.prom_labels, amqp_success="false").inc()
                    raise



import types

from contextlib import nullcontext as does_not_raise
from unittest import mock

import pytest

from prometheus_client import REGISTRY
from pyramid.response import Response

from baseplate.frameworks.pyramid import _make_baseplate_tween
from baseplate.frameworks.pyramid import ACTIVE_REQUESTS
from baseplate.frameworks.pyramid import BaseplateConfigurator
from baseplate.frameworks.pyramid import REQUEST_LATENCY
from baseplate.frameworks.pyramid import REQUEST_SIZE
from baseplate.frameworks.pyramid import REQUESTS_TOTAL
from baseplate.frameworks.pyramid import RESPONSE_SIZE


class TestPyramidHttpServerIntegrationPrometheus:
    def setup(self):
        ACTIVE_REQUESTS.clear()
        REQUESTS_TOTAL.clear()
        REQUEST_LATENCY.clear()
        REQUEST_SIZE.clear()
        RESPONSE_SIZE.clear()

    @pytest.mark.parametrize(
        "response",
        [
            Response("ok"),
            Response("page not found", status=404),
            Response("service unavailable", status=503),
            Exception("some generic exception"),
        ],
    )
    def test_http_server_metric_collection_method_with_request_pattern(self, response):
        status_code = ""
        http_success = "false"
        expectation = does_not_raise()
        if not isinstance(response, Exception):
            status_code = str(response.status_code)
            if status_code == "200":
                http_success = "true"
        else:
            expectation = pytest.raises(type(response))

        prom_labels = {
            "http_method": "get",
            "http_endpoint": "route_pattern",
        }

        handler = mock.MagicMock(return_value=response)
        registry = mock.MagicMock()
        request = mock.MagicMock(content_length=42, method="GET")
        request.matched_route.name = "route"
        request.matched_route.pattern = "route_pattern"
        event = mock.MagicMock(request=request)
        bpConfigurator = BaseplateConfigurator(mock.MagicMock())

        mock_manager = mock.Mock()
        with mock.patch.object(
            ACTIVE_REQUESTS.labels(**prom_labels),
            "inc",
            wraps=ACTIVE_REQUESTS.labels(**prom_labels).inc,
        ) as active_inc_spy_method:
            mock_manager.attach_mock(active_inc_spy_method, "inc")
            with mock.patch.object(
                ACTIVE_REQUESTS.labels(**prom_labels),
                "dec",
                wraps=ACTIVE_REQUESTS.labels(**prom_labels).dec,
            ) as active_dec_spy_method:
                mock_manager.attach_mock(active_dec_spy_method, "dec")
                with mock.patch.object(
                    REQUEST_SIZE.labels(**prom_labels, http_success=http_success),
                    "observe",
                    wraps=REQUEST_SIZE.labels(**prom_labels, http_success=http_success).observe,
                ) as request_size_spy_method:
                    mock_manager.attach_mock(request_size_spy_method, "request_observe")
                    with mock.patch.object(
                        RESPONSE_SIZE.labels(**prom_labels, http_success=http_success),
                        "observe",
                        wraps=RESPONSE_SIZE.labels(
                            **prom_labels, http_success=http_success
                        ).observe,
                    ) as response_size_spy_method:
                        mock_manager.attach_mock(response_size_spy_method, "response_observe")
                        with expectation:
                            bpConfigurator._on_new_request(event)
                            _make_baseplate_tween(handler=handler, _registry=registry)(request)

        assert (
            REGISTRY.get_sample_value(
                "http_server_requests_total",
                {**prom_labels, "http_response_code": status_code, "http_success": http_success},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(
                "http_server_latency_seconds_bucket",
                {**prom_labels, "http_success": http_success, "le": "+Inf"},
            )
            == 1
        )
        assert REGISTRY.get_sample_value("http_server_active_requests", prom_labels) == 0
        expected_calls = [
            mock.call.inc(),  # ensures we first increase number of active requests
            mock.call.dec(),
            mock.call.request_observe(42),
        ]
        if not isinstance(response, Exception):
            expected_calls.append(mock.call.response_observe(response.content_length))
        assert mock_manager.mock_calls == expected_calls

        assert (
            REGISTRY.get_sample_value(
                "http_server_request_size_bytes_bucket",
                {**prom_labels, "http_success": http_success, "le": "+Inf"},
            )
            == 1
        )
        expected_response_size_count = 0 if isinstance(response, Exception) else 1
        assert (
            REGISTRY.get_sample_value(
                "http_server_response_size_bytes_bucket",
                {**prom_labels, "http_success": http_success, "le": "+Inf"},
            )
            == expected_response_size_count
        )

    @pytest.mark.parametrize(
        "response",
        [
            Response("ok"),
            Response("page not found", status=404),
            Response("service unavailable", status=503),
            Exception("some generic exception"),
        ],
    )
    def test_http_server_metric_collection_method_no_pattern(self, response):
        status_code = ""
        http_success = "false"
        expectation = does_not_raise()
        if not isinstance(response, Exception):
            status_code = str(response.status_code)
            if status_code == "200":
                http_success = "true"
        else:
            expectation = pytest.raises(type(response))

        prom_labels = {
            "http_method": "get",
            "http_endpoint": "route_name",
        }

        handler = mock.MagicMock(return_value=response)
        registry = mock.MagicMock()
        request = mock.MagicMock(content_length=42, method="GET")
        request.matched_route = types.SimpleNamespace(name="route_name")
        event = mock.MagicMock(request=request)
        bpConfigurator = BaseplateConfigurator(mock.MagicMock())

        mock_manager = mock.Mock()
        with mock.patch.object(
            ACTIVE_REQUESTS.labels(**prom_labels),
            "inc",
            wraps=ACTIVE_REQUESTS.labels(**prom_labels).inc,
        ) as active_inc_spy_method:
            mock_manager.attach_mock(active_inc_spy_method, "inc")
            with mock.patch.object(
                ACTIVE_REQUESTS.labels(**prom_labels),
                "dec",
                wraps=ACTIVE_REQUESTS.labels(**prom_labels).dec,
            ) as active_dec_spy_method:
                mock_manager.attach_mock(active_dec_spy_method, "dec")
                with mock.patch.object(
                    REQUEST_SIZE.labels(**prom_labels, http_success=http_success),
                    "observe",
                    wraps=REQUEST_SIZE.labels(**prom_labels, http_success=http_success).observe,
                ) as request_size_spy_method:
                    mock_manager.attach_mock(request_size_spy_method, "request_observe")
                    with mock.patch.object(
                        RESPONSE_SIZE.labels(**prom_labels, http_success=http_success),
                        "observe",
                        wraps=RESPONSE_SIZE.labels(
                            **prom_labels, http_success=http_success
                        ).observe,
                    ) as response_size_spy_method:
                        mock_manager.attach_mock(response_size_spy_method, "response_observe")
                        with expectation:
                            bpConfigurator._on_new_request(event)
                            _make_baseplate_tween(handler=handler, _registry=registry)(request)

        assert (
            REGISTRY.get_sample_value(
                "http_server_requests_total",
                {**prom_labels, "http_response_code": status_code, "http_success": http_success},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(
                "http_server_latency_seconds_bucket",
                {**prom_labels, "http_success": http_success, "le": "+Inf"},
            )
            == 1
        )
        assert REGISTRY.get_sample_value("http_server_active_requests", prom_labels) == 0
        expected_calls = [
            mock.call.inc(),  # ensures we first increase number of active requests
            mock.call.dec(),
            mock.call.request_observe(42),
        ]
        if not isinstance(response, Exception):
            expected_calls.append(mock.call.response_observe(response.content_length))
        assert mock_manager.mock_calls == expected_calls

        assert (
            REGISTRY.get_sample_value(
                "http_server_request_size_bytes_bucket",
                {**prom_labels, "http_success": http_success, "le": "+Inf"},
            )
            == 1
        )
        expected_response_size_count = 0 if isinstance(response, Exception) else 1
        assert (
            REGISTRY.get_sample_value(
                "http_server_response_size_bytes_bucket",
                {**prom_labels, "http_success": http_success, "le": "+Inf"},
            )
            == expected_response_size_count
        )



from typing import Dict

from baseplate.lib import config


# default_latency_buckets creates the default bucket values for time based histogram metrics.
# we want this to match the baseplate.go default_buckets
# bp.go v0 ref: https://github.com/reddit/baseplate.go/blob/master/prometheusbp/metrics.go.
# bp.go v2 ref: https://XXXXXXXXX/baseplate/blob/main/metricsbp/metricsbp.go

default_latency_buckets = [
    0.000100,  # 100us
    0.000500,  # 500us
    0.001000,  # 1ms
    0.002500,  # 2.5ms
    0.005000,  # 5ms
    0.010000,  # 10ms
    0.025000,  # 25ms
    0.050000,  # 50ms
    0.100000,  # 100ms
    0.250000,  # 250ms
    0.500000,  # 500ms
    1.000000,  # 1s
    5.000000,  # 5s
    15.000000,  # 15s (fastly timeout)
    30.000000,  # 30s
]

# Default buckets for size base histograms, from <=8 bytes to 4mB in 20
# increments (8*2^i).  Larger requests go in the +Inf bucket.
default_size_start = 8
default_size_factor = 2
default_size_count = 20
default_size_buckets = [
    default_size_start * default_size_factor ** i for i in range(default_size_count)
]


def getHTTPSuccessLabel(httpStatusCode: int) -> str:
    """
    The HTTP success label is "true" if the status code is 2xx or 3xx, "false" otherwise.
    """
    return str(200 <= httpStatusCode < 400).lower()


def is_metrics_enabled(raw_config: Dict[str, str]) -> bool:
    cfg = config.parse_config(
        raw_config,
        {
            "metrics": {
                "enabled": config.Optional(config.Boolean),
            }
        },
    )

    if cfg.metrics.enabled is not None:
        return cfg.metrics.enabled

    return True



"""Memcache serialization/deserialization (and compression) helper methods.

Memcached can only store strings, so to store arbitrary objects we need to
serialize them to strings and be able to deserialize them back to their
original form.

New services should use dump_and_compress() and decompress_and_load().

Services that need to read and write to the same memcache instances as r2
should use pickle_and_compress() and decompress_and_unpickle().

"""
import json
import logging
import pickle
import zlib

from typing import Any
from typing import Callable
from typing import Tuple


class Flags:
    """Memcached client flags.

    Flags are an arbitrary 16-bit unsigned integer that the memcache server
    stores along with the data and sends back when the item is retrieved.
    Clients may use this as a bit field to store data-specific information;
    this field is opaque to the server.

    """

    JSON = 1 << 0
    INTEGER = 1 << 1
    LONG = 1 << 2
    ZLIB = 1 << 3


def decompress_and_load(  # pylint: disable=unused-argument
    key: str, serialized: bytes, flags: int
) -> Any:
    """Deserialize data.

    This should be paired with
    :py:func:`~baseplate.clients.memcache.lib.make_dump_and_compress_fn`.

    :param key: the memcached key.
    :param serialized: the serialized object returned from memcached.
    :param flags: value stored and returned from memcached for the client
        to use to indicate how the value was serialized.
    :returns: The deserialized value.

    """
    if flags & Flags.ZLIB:
        serialized = zlib.decompress(serialized)
        flags ^= Flags.ZLIB

    if flags == 0:
        return serialized

    if flags in (Flags.INTEGER, Flags.LONG):
        # python3 doesn't have a long integer type, so all integers are written
        # with Flags.INTEGER. This means that a value written by a python3
        # client can exceed the maximum integer value (sys.maxint). This
        # appears to be ok--python2 will automatically convert to long if the
        # value is too large.
        return int(serialized)

    if flags == Flags.JSON:
        try:
            return json.loads(serialized)
        except ValueError:
            logging.info("json error", exc_info=True)
            return None

    logging.info("unrecognized flags")
    return serialized


def make_dump_and_compress_fn(
    min_compress_length: int = 0, compress_level: int = 1
) -> Callable[[str, Any], Tuple[bytes, int]]:
    """Make a serializer.

    This should be paired with
    :py:func:`~baseplate.clients.memcache.lib.decompress_and_load`.

    The resulting method is a chain of :py:func:`json.loads` and ``zlib``
    compression. Values that are not JSON serializable will result in a
    :py:exc:`TypeError`.

    :param min_compress_length: the minimum serialized string length to
        enable zlib compression. 0 disables compression.
    :param compress_level: zlib compression level. 0 disables compression
        and 9 is the maximum value.
    :returns: The serializer.

    """
    assert min_compress_length >= 0
    assert 0 <= compress_level <= 9

    def dump_and_compress(  # pylint: disable=unused-argument
        key: str, value: Any
    ) -> Tuple[bytes, int]:
        """Serialize a Python object in a way compatible with decompress_and_load().

        :param key: the memcached key.
        :param value: python object to be serialized and set to memcached.
        :returns: value serialized as str, flags int.
        :raises ValueError: if `value` is not JSON serializable

        """
        if isinstance(value, str):
            serialized = value.encode("utf8")
            flags = 0
        elif isinstance(value, bytes):
            serialized = value
            flags = 0
        elif isinstance(value, int):
            serialized = f"{value:d}".encode()
            flags = Flags.INTEGER
        else:
            # NOTE: json.dumps raises ValueError if `value` is not serializable
            serialized = json.dumps(value).encode("utf8")
            flags = Flags.JSON

        if compress_level and min_compress_length and len(serialized) > min_compress_length:
            serialized = zlib.compress(serialized, compress_level)
            flags |= Flags.ZLIB

        return serialized, flags

    return dump_and_compress


class PickleFlags:
    """Memcached client flags.

    Flags are an arbitrary 16-bit unsigned integer that the memcache server
    stores along with the data and sends back when the item is retrieved.
    Clients may use this as a bit field to store data-specific information;
    this field is opaque to the server.

    """

    PICKLE = 1 << 0
    INTEGER = 1 << 1
    LONG = 1 << 2
    ZLIB = 1 << 3


def decompress_and_unpickle(  # pylint: disable=unused-argument
    key: str, serialized: bytes, flags: int
) -> Any:
    """Deserialize data stored by ``pylibmc``.

    .. warning:: This should only be used when sharing caches with applications
        using ``pylibmc`` (like r2).  New applications should use the safer and
        future proofed
        :py:func:`~baseplate.clients.memcache.lib.decompress_and_load`.

    :param key: the memcached key.
    :param serialized: the serialized object returned from memcached.
    :param flags: value stored and returned from memcached for the client
        to use to indicate how the value was serialized.
    :returns: the deserialized value.

    """
    if flags & PickleFlags.ZLIB:
        serialized = zlib.decompress(serialized)
        flags ^= PickleFlags.ZLIB

    if flags == 0:
        return serialized

    if flags in (PickleFlags.INTEGER, PickleFlags.LONG):
        # python3 doesn't have a long integer type, so all integers are written
        # with PickleFlags.INTEGER. This means that a value written by a python3
        # client can exceed the maximum integer value (sys.maxint). This
        # appears to be ok--python2 will automatically convert to long if the
        # value is too large.
        return int(serialized)

    if flags == PickleFlags.PICKLE:
        try:
            return pickle.loads(serialized)
        except Exception:
            logging.info("Pickle error", exc_info=True)
            return None

    logging.info("unrecognized flags")
    return serialized


def make_pickle_and_compress_fn(
    min_compress_length: int = 0, compress_level: int = 1
) -> Callable[[str, Any], Tuple[bytes, int]]:
    """Make a serializer compatible with ``pylibmc`` readers.

    The resulting method is a chain of :py:func:`pickle.dumps` and ``zlib``
    compression. This should be paired with
    :py:func:`~baseplate.clients.memcache.lib.decompress_and_unpickle`.

    .. warning:: This should only be used when sharing caches with applications
        using ``pylibmc`` (like r2).  New applications should use the safer and
        future proofed
        :py:func:`~baseplate.clients.memcache.lib.make_dump_and_compress_fn`.

    :param min_compress_length: the minimum serialized string length to
        enable zlib compression. 0 disables compression.
    :param compress_level: zlib compression level. 0 disables compression
        and 9 is the maximum value.
    :returns: the serializer method.

    """
    assert min_compress_length >= 0
    assert 0 <= compress_level <= 9

    def pickle_and_compress(  # pylint: disable=unused-argument
        key: str, value: Any
    ) -> Tuple[bytes, int]:
        """Serialize a Python object in a way compatible with decompress_and_unpickle().

        :param key: the memcached key.
        :param value: python object to be serialized and set to memcached.
        :returns: value serialized as str, flags int.

        """
        if isinstance(value, str):
            serialized = value.encode("utf8")
            flags = 0
        elif isinstance(value, bytes):
            serialized = value
            flags = 0
        elif isinstance(value, int):
            serialized = f"{value:d}".encode()
            flags = PickleFlags.INTEGER
        else:
            # use protocol 2 which is the highest value supported by python2
            serialized = pickle.dumps(value, protocol=2)
            flags = PickleFlags.PICKLE

        if compress_level and min_compress_length and len(serialized) > min_compress_length:
            serialized = zlib.compress(serialized, compress_level)
            flags |= PickleFlags.ZLIB

        return serialized, flags

    return pickle_and_compress



import unittest

from io import StringIO
from unittest import mock

from baseplate.lib import service_discovery
from baseplate.lib.file_watcher import FileWatcher
from baseplate.lib.file_watcher import WatchedFileNotAvailableError


TEST_INVENTORY_ONE = """\
[
    {
        "haproxy_server_options": null,
        "host": "10.0.1.2",
        "id": 205,
        "labels": null,
        "name": "i-258fc8b6",
        "port": 9090,
        "weight": null
    }
]
"""

TEST_INVENTORY_TWO = """\
[
    {
        "haproxy_server_options": null,
        "host": "10.0.1.2",
        "id": 205,
        "labels": null,
        "name": "i-258fc8b6",
        "port": 9090,
        "weight": 1
    },
    {
        "haproxy_server_options": null,
        "host": "10.0.1.3",
        "id": 215,
        "labels": null,
        "name": "i-650b6a6b",
        "port": 9090,
        "weight": 2
    },
    {
        "haproxy_server_options": null,
        "host": "10.0.1.4",
        "id": 216,
        "labels": null,
        "name": "i-deadbeef",
        "port": 9090,
        "weight": 1
    }
]
"""


class ServiceInventoryTests(unittest.TestCase):
    def setUp(self):
        self.mock_filewatcher = mock.Mock(spec=FileWatcher)
        self.inventory = service_discovery.ServiceInventory("/whatever")
        self.inventory._filewatcher = self.mock_filewatcher

    def _set_inventory_contents(self, text):
        parsed = service_discovery._parse(StringIO(text))
        self.mock_filewatcher.get_data.return_value = parsed

    def test_load_backends(self):
        self._set_inventory_contents(TEST_INVENTORY_ONE)

        backends = self.inventory.get_backends()
        self.assertEqual(len(backends), 1)
        self.assertEqual(backends[0].id, 205)
        self.assertEqual(backends[0].name, "i-258fc8b6")
        self.assertEqual(backends[0].endpoint.address.host, "10.0.1.2")
        self.assertEqual(backends[0].endpoint.address.port, 9090)
        self.assertEqual(backends[0].weight, 1)

        self._set_inventory_contents(TEST_INVENTORY_TWO)

        backends = self.inventory.get_backends()
        self.assertEqual(len(backends), 3)

    def test_single_get(self):
        self._set_inventory_contents(TEST_INVENTORY_ONE)
        backend = self.inventory.get_backend()
        self.assertEqual(backend.id, 205)

    def test_no_backends_available(self):
        self.mock_filewatcher.get_data.side_effect = WatchedFileNotAvailableError("", None)
        with self.assertRaises(service_discovery.NoBackendsAvailableError):
            self.inventory.get_backend()
        self.assertEqual(self.inventory.get_backends(), [])
        self.mock_filewatcher.get_data.side_effect = None

        self._set_inventory_contents("[]")
        with self.assertRaises(service_discovery.NoBackendsAvailableError):
            self.inventory.get_backend()



import logging
import random

from datetime import timedelta
from time import perf_counter
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import rediscluster

from redis import RedisError
from rediscluster.pipeline import ClusterPipeline

from baseplate import Span
from baseplate.clients import ContextFactory
from baseplate.clients.redis import ACTIVE_REQUESTS
from baseplate.clients.redis import LATENCY_SECONDS
from baseplate.clients.redis import MAX_CONNECTIONS
from baseplate.clients.redis import OPEN_CONNECTIONS
from baseplate.clients.redis import PROM_LABELS_PREFIX
from baseplate.clients.redis import REQUESTS_TOTAL
from baseplate.lib import config
from baseplate.lib import metrics

logger = logging.getLogger(__name__)
randomizer = random.SystemRandom()


# Read commands that take a single key as their first parameter
SINGLE_KEY_READ_COMMANDS = frozenset(
    [
        "BITCOUNT",
        "BITPOS",
        "GEODIST",
        "GEOHASH",
        "GEOPOS",
        "GEORADIUS",
        "GEORADIUSBYMEMBER",
        "GET",
        "GETBIT",
        "GETRANGE",
        "HEXISTS",
        "HGET",
        "HGETALL",
        "HKEYS",
        "HLEN",
        "HMGET",
        "HSTRLEN",
        "HVALS",
        "LINDEX",
        "LLEN",
        "LRANGE",
        "PTTL",
        "SCARD",
        "SISMEMBER",
        "SMEMBERS",
        "SRANDMEMBER",
        "STRLEN",
        "TTL",
        "ZCARD",
        "ZCOUNT",
        "ZRANGE",
        "ZSCORE",
    ]
)

# Read commands that take a list of keys as parameters
MULTI_KEY_READ_COMMANDS = frozenset(["EXISTS", "MGET", "SDIFF", "SINTER", "SUNION"])

# Write commands that take a single key as their first parameter.
SINGLE_KEY_WRITE_COMMANDS = frozenset(
    [
        "EXPIRE",
        "EXPIREAT",
        "HINCRBY",
        "HINCRBYFLOAT",
        "HDEL",
        "HMSET",
        "HSET",
        "HSETNX",
        "LPUSH",
        "LREM",
        "LPOP",
        "LSET",
        "LTRIM",
        "RPOP",
        "RPUSH",
        "SADD",
        "SET",
        "SETNX",
        "SPOP",
        "SREM",
        "ZADD",
        "ZINCRBY",
        "ZPOPMAX",
        "ZPOPMIN",
        "ZREM",
        "ZREMRANGEBYLEX",
        "ZREMRANGEBYRANK",
        "ZREMRANGEBYSCORE",
    ]
)

# Write commands that take a list of keys as argument
MULTI_KEY_WRITE_COMMANDS = frozenset(["DEL"])

# These are a special case of multi-key write commands that take arguments in the form
#  of key value [key value ...]
MULTI_KEY_BATCH_WRITE_COMMANDS = frozenset(["MSET", "MSETNX"])


class HotKeyTracker:
    """
    HotKeyTracker can be used to help identify hot keys within Redis.

    Helper class that can be used to track our key usage and identify hot keys within
    Redis. Whenever we send a read command to Redis we have a (very low but configurable)
    chance to increase a counter associated with that key in Redis. Over time this should
    allow us to find keys that are disproportionaly represented by querying the sorted
    set "baseplate-hot-key-tracker-reads" in Redis. A same sorted set by the name of
    "baseplate-hot-key-tracker-writes" will be used to track write frequency.

    Both read and writes tracking have different configurable percentages, which means
    we can enable tracking for reads without enabling it for writes or have different
    percentages for them, which is useful when the number of reads is much higher than
    the number of writes to a cluster.

    This feature can be turned off by setting the tracking percentage to zero, and should
    probably only be enabled if we're actively debugging an issue or looking for a regression.

    The "baseplate-hot-key-tracker-reads" will have a TTL of 24 hours to ensure that
    older key counts don't interfere with new debugging sessions. This means that the
    sorted set and its contents will disappear in 24 hours after this feature is disabled
    and we stopped writing to it.
    """

    def __init__(
        self,
        redis_client: rediscluster.RedisCluster,
        track_reads_sample_rate: float,
        track_writes_sample_rate: float,
    ):
        self.redis_client = redis_client
        self.track_reads_sample_rate = track_reads_sample_rate
        self.track_writes_sample_rate = track_writes_sample_rate

        self.reads_sorted_set_name = "baseplate-hot-key-tracker-reads"
        self.writes_sorted_set_name = "baseplate-hot-key-tracker-writes"

    def should_track_key_reads(self) -> bool:
        return randomizer.random() < self.track_reads_sample_rate

    def should_track_key_writes(self) -> bool:
        return randomizer.random() < self.track_writes_sample_rate

    def increment_keys_read_counter(self, key_list: List[str], ignore_errors: bool = True) -> None:
        self._increment_hot_key_counter(key_list, self.reads_sorted_set_name, ignore_errors)

    def increment_keys_written_counter(
        self, key_list: List[str], ignore_errors: bool = True
    ) -> None:
        self._increment_hot_key_counter(key_list, self.writes_sorted_set_name, ignore_errors)

    def _increment_hot_key_counter(
        self, key_list: List[str], set_name: str, ignore_errors: bool = True
    ) -> None:
        if len(key_list) == 0:
            return

        try:
            with self.redis_client.pipeline(set_name) as pipe:
                for key in key_list:
                    pipe.zincrby(set_name, 1, key)
                # Reset the TTL for the sorted set
                pipe.expire(set_name, timedelta(hours=24))
                pipe.execute()
        except Exception as e:
            # We don't want to disrupt this request even if key tracking fails, so just
            # log it.
            logger.exception(e)
            if not ignore_errors:
                raise

    def maybe_track_key_usage(self, args: List[str]) -> None:
        """Probabilistically track usage of the keys in this command.

        If we have enabled key usage tracing *and* this command is withing the
        percentage of commands we want to track, then write it to a sorted set
        so we can keep track of the most accessed keys.
        """
        if len(args) == 0:
            return

        command = args[0]

        if self.should_track_key_reads():
            if command in SINGLE_KEY_READ_COMMANDS:
                self.increment_keys_read_counter([args[1]])
            elif command in MULTI_KEY_READ_COMMANDS:
                self.increment_keys_read_counter(args[1:])

        if self.should_track_key_writes():
            if command in SINGLE_KEY_WRITE_COMMANDS:
                self.increment_keys_written_counter([args[1]])
            elif command in MULTI_KEY_WRITE_COMMANDS:
                self.increment_keys_written_counter(args[1:])
            elif command in MULTI_KEY_BATCH_WRITE_COMMANDS:
                # These commands follow key value [key value...] format
                self.increment_keys_written_counter(args[1::2])


# We want to be able to combine blocking behaviour with the ability to read from replicas
# Unfortunately this is not provide as-is so we combine two connection pool classes to provide
# the desired behaviour.
class ClusterWithReadReplicasBlockingConnectionPool(rediscluster.ClusterBlockingConnectionPool):
    # pylint: disable=arguments-differ
    def get_node_by_slot(self, slot: int, read_command: bool = False) -> Dict[str, Any]:
        """Get a node from the slot.

        If the command is a read command we'll try to return a random node.
        If there are no replicas or this isn't a read command we'll return the primary.
        """
        if read_command:
            return random.choice(self.nodes.slots[slot])

        # This isn't a read command, so return the primary (first node)
        return self.nodes.slots[slot][0]


def cluster_pool_from_config(
    app_config: config.RawConfig, prefix: str = "rediscluster.", **kwargs: Any
) -> rediscluster.ClusterConnectionPool:
    """Make a ClusterConnectionPool from a configuration dictionary.

    The keys useful to :py:func:`cluster_pool_from_config` should be prefixed, e.g.
    ``rediscluster.url``, ``rediscluster.max_connections``, etc. The ``prefix`` argument
    specifies the prefix used to filter keys.  Each key is mapped to a
    corresponding keyword argument on the :py:class:`rediscluster.ClusterConnectionPool`
    constructor.

    Supported keys:

    * ``url`` (required): a URL like ``redis://localhost/0``.
    * ``max_connections``: an integer maximum number of connections in the pool
    * ``max_connections_per_node``: Boolean, whether max_connections should be applied
        globally (False) or per node (True).
    * ``skip_full_coverage_check``: Skips the check of cluster-require-full-coverage
      config, useful for clusters without the CONFIG command (like aws)
    * ``nodemanager_follow_cluster``: Tell the node manager to reuse the last set of
      nodes it was operating on when intializing.
    * ``read_from_replicas``: (Boolean) Whether the client should send all read queries to
        replicas instead of just the primary
    * ``timeout``: . e.g. ``200 milliseconds`` (:py:func:`~baseplate.lib.config.Timespan`).
        How long to wait for a connection to become available.  Additionally, will set
        ``socket_connect_timeout`` and ``socket_timeout`` if they're not set explicitly.
    * ``socket_connect_timeout``: e.g. ``200 milliseconds`` (:py:func:`~baseplate.lib.config.Timespan`)
        How long to wait for sockets to connect.
    * ``socket_timeout``: e.g. ``200 milliseconds`` (:py:func:`~baseplate.lib.config.Timespan`)
        How long to wait for socket operations.
    * ``track_key_reads_sample_rate``: If greater than zero, which percentage of requests will
        be inspected to keep track of hot key usage within Redis when reading.
        Every command inspected will result in a write to a sorted set
        (baseplate-hot-key-tracker-reads) for tracking.
    * ``track_key_writes_sample_rate``: If greater than zero, which percentage of requests will
        be inspected to keep track of hot key usage within Redis when writing.
        Every command inspected will result in a write to a sorted set
        (baseplate-hot-key-tracker-reads) for tracking.

    """
    assert prefix.endswith(".")

    parser = config.SpecParser(
        {
            "url": config.String,
            "max_connections": config.Optional(config.Integer, default=50),
            "max_connections_per_node": config.Optional(config.Boolean, default=False),
            "socket_connect_timeout": config.Optional(config.Timespan, default=None),
            "socket_timeout": config.Optional(config.Timespan, default=None),
            "timeout": config.Optional(config.Timespan, default=None),
            "read_from_replicas": config.Optional(config.Boolean, default=True),
            "skip_full_coverage_check": config.Optional(config.Boolean, default=True),
            "nodemanager_follow_cluster": config.Optional(config.Boolean, default=None),
            "decode_responses": config.Optional(config.Boolean, default=True),
            "track_key_reads_sample_rate": config.Optional(config.Float, default=0),
            "track_key_writes_sample_rate": config.Optional(config.Float, default=0),
        }
    )

    options = parser.parse(prefix[:-1], app_config)

    # We're explicitly setting a default here because of https://github.com/Grokzen/redis-py-cluster/issues/435
    kwargs.setdefault("max_connections", options.max_connections)

    kwargs.setdefault("decode_responses", options.decode_responses)

    if options.nodemanager_follow_cluster is not None:
        kwargs.setdefault("nodemanager_follow_cluster", options.nodemanager_follow_cluster)
    if options.skip_full_coverage_check is not None:
        kwargs.setdefault("skip_full_coverage_check", options.skip_full_coverage_check)
    if options.socket_connect_timeout is not None:
        kwargs.setdefault("socket_connect_timeout", options.socket_connect_timeout.total_seconds())
    if options.socket_timeout is not None:
        kwargs.setdefault("socket_timeout", options.socket_timeout.total_seconds())
    if options.timeout is not None:
        kwargs.setdefault("timeout", options.timeout.total_seconds())
        # set socket_connection_timeout and socket_timeout if not set already
        kwargs.setdefault("socket_connect_timeout", options.timeout.total_seconds())
        kwargs.setdefault("socket_timeout", options.timeout.total_seconds())

    if options.read_from_replicas:
        connection_pool = ClusterWithReadReplicasBlockingConnectionPool.from_url(
            options.url, **kwargs
        )
    else:
        connection_pool = rediscluster.ClusterBlockingConnectionPool.from_url(options.url, **kwargs)

    connection_pool.track_key_reads_sample_rate = options.track_key_reads_sample_rate
    connection_pool.track_key_writes_sample_rate = options.track_key_writes_sample_rate

    connection_pool.read_from_replicas = options.read_from_replicas
    connection_pool.skip_full_coverage_check = options.skip_full_coverage_check

    return connection_pool


class ClusterRedisClient(config.Parser):
    """Configure a clustered Redis client.

    :param redis_client_name: The name of this Redis client. Prefer to use `client_name`
    keyword argument to the `pool_from_config` function. Use this if your using a Redis
    host or proxy that doesn't support the `CLIENT SETNAME` function.

    This is meant to be used with
    :py:meth:`baseplate.Baseplate.configure_context`.
    See :py:func:`cluster_pool_from_config` for available configuration settings.
    """

    def __init__(self, redis_client_name: str = "", **kwargs: Any):
        # This is for backwards compatibility. Originally we asked clients to
        # set the `client_name` attribute to get the `redis_client_name`
        # tag to appear on Prometheus metrics. Unfortunately this broke clients
        # that use a proxy to connect to Redis.
        # See: https://github.com/redis/redis-py/issues/2384/
        client_name = redis_client_name
        if client_name == "":
            client_name = kwargs.get("client_name", "")

        self.kwargs = kwargs
        self.redis_client_name = client_name

    def parse(self, key_path: str, raw_config: config.RawConfig) -> "ClusterRedisContextFactory":
        connection_pool = cluster_pool_from_config(raw_config, f"{key_path}.", **self.kwargs)
        return ClusterRedisContextFactory(
            connection_pool, key_path, redis_client_name=self.redis_client_name
        )


class ClusterRedisContextFactory(ContextFactory):
    """Cluster Redis client context factory.

    This factory will attach a
    :py:class:`~baseplate.clients.redis.MonitoredRedisClusterConnection` to an
    attribute on the :py:class:`~baseplate.RequestContext`. When Redis commands
    are executed via this connection object, they will use connections from the
    provided :py:class:`rediscluster.ClusterConnectionPool` and automatically record
    diagnostic information.
    :param connection_pool: A connection pool.
    """

    def __init__(
        self,
        connection_pool: rediscluster.ClusterConnectionPool,
        name: str = "redis",
        redis_client_name: str = "",
    ):
        self.connection_pool = connection_pool
        self.name = name
        self.redis_client_name = redis_client_name

    def report_runtime_metrics(self, batch: metrics.Client) -> None:
        if not isinstance(self.connection_pool, rediscluster.ClusterBlockingConnectionPool):
            return

        size = self.connection_pool.max_connections
        open_connections_num = len(self.connection_pool._connections)
        MAX_CONNECTIONS.labels(self.name).set(size)
        OPEN_CONNECTIONS.labels(self.name).set(open_connections_num)

        batch.gauge("pool.size").replace(size)
        batch.gauge("pool.open_connections").replace(open_connections_num)

    def make_object_for_context(self, name: str, span: Span) -> "MonitoredRedisClusterConnection":
        return MonitoredRedisClusterConnection(
            name,
            span,
            self.connection_pool,
            getattr(self.connection_pool, "track_key_reads_sample_rate", 0),
            getattr(self.connection_pool, "track_key_writes_sample_rate", 0),
            self.redis_client_name,
        )


class MonitoredRedisClusterConnection(rediscluster.RedisCluster):
    """Cluster Redis connection that collects diagnostic information.

    This connection acts like :py:class:`rediscluster.Redis` except that all
    operations are automatically wrapped with diagnostic collection.
    The interface is the same as that class except for the
    :py:meth:`~baseplate.clients.redis.MonitoredRedisClusterConnection.pipeline`
    method.
    """

    def __init__(
        self,
        context_name: str,
        server_span: Span,
        connection_pool: rediscluster.ClusterConnectionPool,
        track_key_reads_sample_rate: float = 0,
        track_key_writes_sample_rate: float = 0,
        redis_client_name: str = "",
    ):
        self.context_name = context_name
        self.server_span = server_span
        self.track_key_reads_sample_rate = track_key_reads_sample_rate
        self.track_key_writes_sample_rate = track_key_writes_sample_rate
        self.hot_key_tracker = HotKeyTracker(
            self, self.track_key_reads_sample_rate, self.track_key_writes_sample_rate
        )
        self.redis_client_name = redis_client_name

        super().__init__(
            connection_pool=connection_pool,
            read_from_replicas=connection_pool.read_from_replicas,
            skip_full_coverage_check=connection_pool.skip_full_coverage_check,
        )

    def execute_command(self, *args: Any, **kwargs: Any) -> Any:
        command = args[0]
        trace_name = f"{self.context_name}.{command}"

        with self.server_span.make_child(trace_name):
            start_time = perf_counter()
            success = "true"
            labels = {
                f"{PROM_LABELS_PREFIX}_command": command,
                f"{PROM_LABELS_PREFIX}_client_name": self.redis_client_name,
                f"{PROM_LABELS_PREFIX}_database": self.connection_pool.connection_kwargs.get(
                    "db", ""
                ),
                f"{PROM_LABELS_PREFIX}_type": "cluster",
            }

            try:
                with ACTIVE_REQUESTS.labels(**labels).track_inprogress():
                    res = super().execute_command(command, *args[1:], **kwargs)
                if isinstance(res, RedisError):
                    success = "false"
            except:  # noqa: E722
                success = "false"
                raise
            finally:
                result_labels = {**labels, f"{PROM_LABELS_PREFIX}_success": success}
                REQUESTS_TOTAL.labels(**result_labels).inc()
                LATENCY_SECONDS.labels(**result_labels).observe(perf_counter() - start_time)

        self.hot_key_tracker.maybe_track_key_usage(list(args))

        return res

    # pylint: disable=arguments-differ
    def pipeline(self, name: str) -> "MonitoredClusterRedisPipeline":
        """Create a pipeline.

        This returns an object on which you can call the standard Redis
        commands. Execution will be deferred until ``execute`` is called. This
        is useful for saving round trips even in a clustered environment .
        :param name: The name to attach to diagnostics for this pipeline.
        """
        return MonitoredClusterRedisPipeline(
            f"{self.context_name}.pipeline_{name}",
            self.server_span,
            self.connection_pool,
            self.response_callbacks,
            read_from_replicas=self.read_from_replicas,
            hot_key_tracker=self.hot_key_tracker,
            redis_client_name=self.redis_client_name,
        )

    # No transaction support in redis-py-cluster
    def transaction(self, *args: Any, **kwargs: Any) -> Any:
        """Not currently implemented."""
        raise NotImplementedError


# pylint: disable=abstract-method
class MonitoredClusterRedisPipeline(ClusterPipeline):
    def __init__(
        self,
        trace_name: str,
        server_span: Span,
        connection_pool: rediscluster.ClusterConnectionPool,
        response_callbacks: Dict,
        hot_key_tracker: Optional[HotKeyTracker],
        redis_client_name: str = "",
        **kwargs: Any,
    ):
        self.trace_name = trace_name
        self.server_span = server_span
        self.hot_key_tracker = hot_key_tracker
        self.redis_client_name = redis_client_name
        super().__init__(connection_pool, response_callbacks, **kwargs)

    def execute_command(self, *args: Any, **kwargs: Any) -> Any:
        res = super().execute_command(*args, **kwargs)

        if self.hot_key_tracker is not None:
            self.hot_key_tracker.maybe_track_key_usage(list(args))

        return res

    # pylint: disable=arguments-differ
    def execute(self, **kwargs: Any) -> Any:
        with self.server_span.make_child(self.trace_name):
            success = "true"
            start_time = perf_counter()
            labels = {
                f"{PROM_LABELS_PREFIX}_command": "pipeline",
                f"{PROM_LABELS_PREFIX}_client_name": self.redis_client_name,
                f"{PROM_LABELS_PREFIX}_database": self.connection_pool.connection_kwargs.get(
                    "db", ""
                ),
                f"{PROM_LABELS_PREFIX}_type": "cluster",
            }

            ACTIVE_REQUESTS.labels(**labels).inc()

            try:
                return super().execute(**kwargs)
            except:  # noqa: E722
                success = "false"
                raise
            finally:
                ACTIVE_REQUESTS.labels(**labels).dec()
                result_labels = {
                    **labels,
                    f"{PROM_LABELS_PREFIX}_success": success,
                }
                REQUESTS_TOTAL.labels(**result_labels).inc()
                LATENCY_SECONDS.labels(**result_labels).observe(perf_counter() - start_time)



"""Utilities for common cryptographic operations.

.. testsetup::

    import datetime
    from baseplate.lib.crypto import make_signature, validate_signature
    from baseplate.lib.secrets import SecretsStore

    secrets = SecretsStore("docs/secrets.json")

.. testcode::

    message = "Hello, world!"

    secret = secrets.get_versioned("some_signing_key")
    signature = make_signature(
        secret, message, max_age=datetime.timedelta(days=1))

    try:
        validate_signature(secret, message, signature)
    except SignatureError:
        print("Oh no, it was invalid!")
    else:
        print("Message was valid!")

.. testoutput::

    Message was valid!


"""
import base64
import binascii
import datetime
import hashlib
import hmac
import struct
import time

from typing import NamedTuple

from baseplate.lib.secrets import VersionedSecret


class SignatureError(Exception):
    """Base class for all message signing related errors."""


class UnreadableSignatureError(SignatureError):
    """Raised when the signature is corrupt or wrongly formatted."""


class IncorrectSignatureError(SignatureError):
    """Raised when the signature is readable but does not match the message."""


class ExpiredSignatureError(SignatureError):
    """Raised when the signature is valid but has expired.

    The ``expiration`` attribute is the time (as seconds since the UNIX epoch)
    at which the signature expired.

    """

    def __init__(self, expiration: int):
        self.expiration = expiration
        super().__init__()


# A signature is a base64 encoded binary blob, comprised of a header and
# digest.
#
# The first byte of the header is a version number indicating what format the
# signature is.
#
# In version 1, the only current version, the header then has two bytes of
# padding, to prevent base64 "=" padding, followed by the expiration time of
# the signature as seconds since the unix epoch. An HMAC-SHA256 digest follows.
_HEADER_FORMAT = struct.Struct("<BxxI")


class SignatureInfo(NamedTuple):
    """Information about a valid signature.

    :ivar version: The version of the packed signature format.
    :ivar expiration: The time, in seconds since the UNIX epoch, at which
        the signature will expire.

    """

    version: int
    expiration: int


def _compute_digest(secret_value: bytes, header: bytes, message: str) -> bytes:
    payload = header + message.encode("utf8")
    digest = hmac.new(secret_value, payload, hashlib.sha256).digest()  # pylint: disable=no-member
    return digest


def make_signature(secret: VersionedSecret, message: str, max_age: datetime.timedelta) -> bytes:
    """Return a signature for the given message.

    To ensure that key rotation works automatically, always fetch the secret
    token from the secret store immediately before use and do not cache / save
    the token anywhere. The ``current`` version of the secret will be used to
    sign the token.

    :param secret: The secret signing key from the secret store.
    :param message: The message to sign.
    :param max_age: The amount of time in the future the signature will be valid for.
    :return: An encoded signature.

    """
    version = 1
    expiration = int(time.time() + max_age.total_seconds())
    header = _HEADER_FORMAT.pack(version, expiration)
    digest = _compute_digest(secret.current, header, message)
    return base64.urlsafe_b64encode(header + digest)


def validate_signature(secret: VersionedSecret, message: str, signature: bytes) -> SignatureInfo:
    """Validate and assert a message's signature is correct.

    If the signature is valid, the function will return normally with a
    :py:class:`SignatureInfo` with some details about the signature.
    Otherwise, an exception will be raised.

    To ensure that key rotation works automatically, always fetch the secret
    token from the secret store immediately before use and do not cache / save
    the token anywhere. All active versions of the secret will be checked when
    validating the signature.

    :param secret: The secret signing key from the secret store.
    :param message: The message payload to validate.
    :param signature: The signature supplied with the message.
    :raises: :py:exc:`UnreadableSignatureError` The signature is corrupt.
    :raises: :py:exc:`IncorrectSignatureError` The digest is incorrect.
    :raises: :py:exc:`ExpiredSignatureError` The signature expired.

    """
    version: int
    expiration: int

    try:
        signature_bytes = base64.urlsafe_b64decode(signature)
        header = signature_bytes[: _HEADER_FORMAT.size]
        signature_digest = signature_bytes[_HEADER_FORMAT.size :]
        version, expiration = _HEADER_FORMAT.unpack(header)
        if version != 1:
            raise ValueError
        if len(signature_digest) != hashlib.sha256().digest_size:  # pylint: disable=no-member
            raise ValueError
    except (struct.error, KeyError, binascii.Error, TypeError, ValueError):
        raise UnreadableSignatureError

    if time.time() > expiration:
        raise ExpiredSignatureError(expiration)

    for secret_value in secret.all_versions:
        digest = _compute_digest(secret_value, header, message)
        if hmac.compare_digest(digest, signature_digest):
            break
    else:
        raise IncorrectSignatureError

    return SignatureInfo(version, expiration)
